cdef class EllipsoidSelector(SelectorObject):
    cdef public np.float64_t vec[3][3]
    cdef public np.float64_t mag[3]
    cdef public np.float64_t center[3]

    def __init__(self, dobj):
        cdef int i
        _ensure_code(dobj.center)
        _ensure_code(dobj._e0)
        _ensure_code(dobj._e1)
        _ensure_code(dobj._e2)
        _ensure_code(dobj._A)
        _ensure_code(dobj._B)
        _ensure_code(dobj._C)
        for i in range(3):
            self.center[i] = dobj.center[i]
            self.vec[0][i] = dobj._e0[i]
            self.vec[1][i] = dobj._e1[i]
            self.vec[2][i] = dobj._e2[i]
        self.mag[0] = dobj._A
        self.mag[1] = dobj._B
        self.mag[2] = dobj._C

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        return self.select_point(pos)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef np.float64_t dot_evec[3]
        cdef np.float64_t dist
        cdef int i, j
        dot_evec[0] = dot_evec[1] = dot_evec[2] = 0
        # Calculate the rotated dot product
        for i in range(3): # axis
            dist = self.periodic_difference(pos[i], self.center[i], i)
            for j in range(3):
                dot_evec[j] += dist * self.vec[j][i]
        dist = 0.0
        for i in range(3):
            dist += (dot_evec[i] * dot_evec[i])/(self.mag[i] * self.mag[i])
        if dist <= 1.0: return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        # this is the sphere selection
        cdef int i
        cdef np.float64_t dist, dist2_max, dist2 = 0
        for i in range(3):
            dist = self.periodic_difference(pos[i], self.center[i], i)
            dist2 += dist * dist
        dist2_max = (self.mag[0] + radius) * (self.mag[0] + radius)
        if dist2 <= dist2_max:
            return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        # This is the sphere selection
        cdef int i
        cdef np.float64_t box_center, relcenter, closest, dist, edge, dist_max
        if left_edge[0] <= self.center[0] <= right_edge[0] and \
           left_edge[1] <= self.center[1] <= right_edge[1] and \
           left_edge[2] <= self.center[2] <= right_edge[2]:
            return 1
        # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
        dist = 0
        for i in range(3):
            box_center = (right_edge[i] + left_edge[i])/2.0
            relcenter = self.periodic_difference(box_center, self.center[i], i)
            edge = right_edge[i] - left_edge[i]
            closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
            dist += closest * closest
        dist_max = self.mag[0] * self.mag[0]
        if dist <= dist_max:
            return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        # This is the sphere selection
        cdef int i
        cdef np.float64_t box_center, relcenter, closest, farthest, cdist, fdist, edge
        if left_edge[0] <= self.center[0] <= right_edge[0] and \
           left_edge[1] <= self.center[1] <= right_edge[1] and \
           left_edge[2] <= self.center[2] <= right_edge[2]:
            fdist = 0
            for i in range(3):
                edge = right_edge[i] - left_edge[i]
                box_center = (right_edge[i] + left_edge[i])/2.0
                relcenter = self.periodic_difference(
                    box_center, self.center[i], i)
                farthest = relcenter + fclip(relcenter, -edge/2.0, edge/2.0)
                fdist += farthest*farthest
                if fdist >= self.mag[0]**2: return 2
            return 1
        # http://www.gamedev.net/topic/335465-is-this-the-simplest-sphere-aabb-collision-test/
        cdist = 0
        fdist = 0
        for i in range(3):
            box_center = (right_edge[i] + left_edge[i])/2.0
            relcenter = self.periodic_difference(box_center, self.center[i], i)
            edge = right_edge[i] - left_edge[i]
            closest = relcenter - fclip(relcenter, -edge/2.0, edge/2.0)
            farthest = relcenter + fclip(relcenter, -edge/2.0, edge/2.0)
            cdist += closest * closest
            fdist += farthest * farthest
            if cdist > self.mag[0]**2: return 0
        if fdist < self.mag[0]**2:
            return 1
        else:
            return 2

    def _hash_vals(self):
        return (("vec[0][0]", self.vec[0][0]),
                ("vec[0][1]", self.vec[0][1]),
                ("vec[0][2]", self.vec[0][2]),
                ("vec[1][0]", self.vec[1][0]),
                ("vec[1][1]", self.vec[1][1]),
                ("vec[1][2]", self.vec[1][2]),
                ("vec[2][0]", self.vec[2][0]),
                ("vec[2][1]", self.vec[2][1]),
                ("vec[2][2]", self.vec[2][2]),
                ("mag[0]", self.mag[0]),
                ("mag[1]", self.mag[1]),
                ("mag[2]", self.mag[2]),
                ("center[0]", self.center[0]),
                ("center[1]", self.center[1]),
                ("center[2]", self.center[2]))

    def _get_state_attnames(self):
        return ("mag", "center", "vec")


ellipsoid_selector = EllipsoidSelector



import base64
import os
from io import BytesIO

import numpy as np

from yt.fields.derived_field import ValidateSpatial
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _astropy


def _make_counts(emin, emax):
    def _counts(field, data):
        e = data[("all", "event_energy")].in_units("keV")
        mask = np.logical_and(e >= emin, e < emax)
        x = data[("all", "event_x")][mask]
        y = data[("all", "event_y")][mask]
        z = np.ones(x.shape)
        pos = np.array([x, y, z]).transpose()
        img = data.deposit(pos, method="count")
        if data.has_field_parameter("sigma"):
            sigma = data.get_field_parameter("sigma")
        else:
            sigma = None
        if sigma is not None and sigma > 0.0:
            kern = _astropy.conv.Gaussian2DKernel(x_stddev=sigma)
            img[:, :, 0] = _astropy.conv.convolve(img[:, :, 0], kern)
        return data.ds.arr(img, "counts/pixel")

    return _counts


def setup_counts_fields(ds, ebounds, ftype="gas"):
    r"""
    Create deposited image fields from X-ray count data in energy bands.

    Parameters
    ----------
    ds : ~yt.data_objects.static_output.Dataset
        The FITS events file dataset to add the counts fields to.
    ebounds : list of tuples
        A list of tuples, one for each field, with (emin, emax) as the
        energy bounds for the image.
    ftype : string, optional
        The field type of the resulting field. Defaults to "gas".

    Examples
    --------
    >>> ds = yt.load("evt.fits")
    >>> ebounds = [(0.1, 2.0), (2.0, 3.0)]
    >>> setup_counts_fields(ds, ebounds)
    """
    for emin, emax in ebounds:
        cfunc = _make_counts(emin, emax)
        fname = f"counts_{emin}-{emax}"
        mylog.info("Creating counts field %s.", fname)
        ds.add_field(
            (ftype, fname),
            sampling_type="cell",
            function=cfunc,
            units="counts/pixel",
            validators=[ValidateSpatial()],
            display_name=f"Counts ({emin}-{emax} keV)",
        )


def create_spectral_slabs(filename, slab_centers, slab_width, **kwargs):
    r"""
    Given a dictionary of spectral slab centers and a width in
    spectral units, extract data from a spectral cube at these slab
    centers and return a `FITSDataset` instance containing the different
    slabs as separate yt fields. Useful for extracting individual
    lines from a spectral cube and separating them out as different fields.

    Requires the SpectralCube (https://spectral-cube.readthedocs.io/en/latest/)
    library.

    All keyword arguments will be passed on to the `FITSDataset` constructor.

    Parameters
    ----------
    filename : string
        The spectral cube FITS file to extract the data from.
    slab_centers : dict of (float, string) tuples or YTQuantities
        The centers of the slabs, where the keys are the names
        of the new fields and the values are (float, string) tuples or
        YTQuantities, specifying a value for each center and its unit.
    slab_width : YTQuantity or (float, string) tuple
        The width of the slab along the spectral axis.

    Examples
    --------
    >>> slab_centers = {
    ...     "13CN": (218.03117, "GHz"),
    ...     "CH3CH2CHO": (218.284256, "GHz"),
    ...     "CH3NH2": (218.40956, "GHz"),
    ... }
    >>> slab_width = (0.05, "GHz")
    >>> ds = create_spectral_slabs(
    ...     "intensity_cube.fits", slab_centers, slab_width, nan_mask=0.0
    ... )
    """
    from spectral_cube import SpectralCube

    from yt.frontends.fits.api import FITSDataset
    from yt.visualization.fits_image import FITSImageData

    cube = SpectralCube.read(filename)
    if not isinstance(slab_width, YTQuantity):
        slab_width = YTQuantity(slab_width[0], slab_width[1])
    slab_data = {}
    field_units = cube.header.get("bunit", "dimensionless")
    for k, v in slab_centers.items():
        if not isinstance(v, YTQuantity):
            slab_center = YTQuantity(v[0], v[1])
        else:
            slab_center = v
        mylog.info("Adding slab field %s at %g %s", k, slab_center.v, slab_center.units)
        slab_lo = (slab_center - 0.5 * slab_width).to_astropy()
        slab_hi = (slab_center + 0.5 * slab_width).to_astropy()
        subcube = cube.spectral_slab(slab_lo, slab_hi)
        slab_data[k] = YTArray(subcube.filled_data[:, :, :], field_units)
    width = subcube.header["naxis3"] * cube.header["cdelt3"]
    w = subcube.wcs.copy()
    w.wcs.crpix[-1] = 0.5
    w.wcs.crval[-1] = -0.5 * width
    fid = FITSImageData(slab_data, wcs=w)
    for hdu in fid:
        hdu.header.pop("RESTFREQ", None)
        hdu.header.pop("RESTFRQ", None)
    ds = FITSDataset(fid, **kwargs)
    return ds


def ds9_region(ds, reg, obj=None, field_parameters=None):
    r"""
    Create a data container from a ds9 region file. Requires the regions
    package (https://astropy-regions.readthedocs.io/) to be installed.

    Parameters
    ----------
    ds : FITSDataset
        The Dataset to create the region from.
    reg : string
        The filename of the ds9 region, or a region string to be parsed.
    obj : data container, optional
        The data container that will be used to create the new region.
        Defaults to ds.all_data.
    field_parameters : dictionary, optional
        A set of field parameters to apply to the region.

    Examples
    --------

    >>> ds = yt.load("m33_hi.fits")
    >>> circle_region = ds9_region(ds, "circle.reg")
    >>> print(circle_region.quantities.extrema("flux"))
    """
    from yt.utilities.on_demand_imports import _astropy, _regions

    Regions = _regions.Regions

    WCS = _astropy.WCS

    from yt.frontends.fits.api import EventsFITSDataset

    if os.path.exists(reg):
        method = Regions.read
    else:
        method = Regions.parse
    r = method(reg, format="ds9").regions[0]

    reg_name = reg
    header = ds.wcs_2d.to_header()
    # The FITS header only contains WCS-related keywords
    header["NAXIS1"] = ds.domain_dimensions[ds.lon_axis]
    header["NAXIS2"] = ds.domain_dimensions[ds.lat_axis]

    pixreg = r.to_pixel(WCS(header))
    mask = pixreg.to_mask().to_image((header["NAXIS1"], header["NAXIS2"])).astype(bool)

    if isinstance(ds, EventsFITSDataset):
        prefix = "event_"
    else:
        prefix = ""

    def _reg_field(field, data):
        i = data[prefix + "xyz"[ds.lon_axis]].d.astype("int") - 1
        j = data[prefix + "xyz"[ds.lat_axis]].d.astype("int") - 1
        new_mask = mask[i, j]
        ret = np.zeros(data[prefix + "x"].shape)
        ret[new_mask] = 1.0
        return ret

    ds.add_field(("gas", reg_name), sampling_type="cell", function=_reg_field)
    if obj is None:
        obj = ds.all_data()
    if field_parameters is not None:
        for k, v in field_parameters.items():
            obj.set_field_parameter(k, v)
    return obj.cut_region([f"obj['{reg_name}'] > 0"])


class PlotWindowWCS:
    r"""
    Use AstroPy's WCSAxes class to plot celestial coordinates on the axes of a
    on-axis PlotWindow plot. See
    http://docs.astropy.org/en/stable/visualization/wcsaxes/ for more details
    on how it works under the hood. This functionality requires a version of
    AstroPy >= 1.3.

    Parameters
    ----------
    pw : on-axis PlotWindow instance
        The PlotWindow instance to add celestial axes to.
    """

    def __init__(self, pw):
        WCSAxes = _astropy.wcsaxes.WCSAxes

        if pw.oblique:
            raise NotImplementedError("WCS axes are not implemented for oblique plots.")
        if not hasattr(pw.ds, "wcs_2d"):
            raise NotImplementedError("WCS axes are not implemented for this dataset.")
        if pw.data_source.axis != pw.ds.spec_axis:
            raise NotImplementedError("WCS axes are not implemented for this axis.")
        self.plots = {}
        self.pw = pw
        for f in pw.plots:
            rect = pw.plots[f]._get_best_layout()[1]
            fig = pw.plots[f].figure
            ax = fig.axes[0]
            wcs_ax = WCSAxes(fig, rect, wcs=pw.ds.wcs_2d, frameon=False)
            fig.add_axes(wcs_ax)
            wcs = pw.ds.wcs_2d.wcs
            xax = pw.ds.coordinates.x_axis[pw.data_source.axis]
            yax = pw.ds.coordinates.y_axis[pw.data_source.axis]
            xlabel = f"{wcs.ctype[xax].split('-')[0]} ({wcs.cunit[xax]})"
            ylabel = f"{wcs.ctype[yax].split('-')[0]} ({wcs.cunit[yax]})"
            fp = pw._font_properties
            wcs_ax.coords[0].set_axislabel(xlabel, fontproperties=fp, minpad=0.5)
            wcs_ax.coords[1].set_axislabel(ylabel, fontproperties=fp, minpad=0.4)
            wcs_ax.coords[0].ticklabels.set_fontproperties(fp)
            wcs_ax.coords[1].ticklabels.set_fontproperties(fp)
            ax.xaxis.set_visible(False)
            ax.yaxis.set_visible(False)
            wcs_ax.set_xlim(pw.xlim[0].value, pw.xlim[1].value)
            wcs_ax.set_ylim(pw.ylim[0].value, pw.ylim[1].value)
            wcs_ax.coords.frame._update_cache = []
            ax.xaxis.set_visible(False)
            ax.yaxis.set_visible(False)
            self.plots[f] = fig

    def keys(self):
        return self.plots.keys()

    def values(self):
        return self.plots.values()

    def items(self):
        return self.plots.items()

    def __getitem__(self, key):
        for k in self.keys():
            if k[1] == key:
                return self.plots[k]

    def show(self):
        return self

    def save(self, name=None, mpl_kwargs=None):
        if mpl_kwargs is None:
            mpl_kwargs = {}
        mpl_kwargs["bbox_inches"] = "tight"
        self.pw.save(name=name, mpl_kwargs=mpl_kwargs)

    def _repr_html_(self):
        from yt.visualization._mpl_imports import FigureCanvasAgg

        ret = ""
        for v in self.plots.values():
            canvas = FigureCanvasAgg(v)
            f = BytesIO()
            canvas.print_figure(f)
            f.seek(0)
            img = base64.b64encode(f.read()).decode()
            ret += (
                r'<img style="max-width:100%%;max-height:100%%;" '
                r'src="data:image/png;base64,%s"><br>' % img
            )
        return ret



# image tests using pytest-mpl
import re
from importlib.metadata import version
from itertools import chain
from typing import Dict

import numpy as np
import numpy.testing as npt
import pytest
from packaging.version import Version

from yt.data_objects.profiles import create_profile
from yt.loaders import load_uniform_grid
from yt.testing import add_noise_fields, fake_amr_ds, fake_particle_ds, fake_random_ds
from yt.visualization.api import (
    LinePlot,
    ParticleProjectionPlot,
    PhasePlot,
    ProfilePlot,
    SlicePlot,
)

MPL_VERSION = Version(version("matplotlib"))


def test_sliceplot_set_unit_and_zlim_order():
    ds = fake_random_ds(16)
    field = ("gas", "density")

    p0 = SlicePlot(ds, "z", field)
    p0.set_unit(field, "kg/m**3")
    p0.set_zlim(field, zmin=0)

    # reversing order of operations
    p1 = SlicePlot(ds, "z", field)
    p1.set_zlim(field, zmin=0)
    p1.set_unit(field, "kg/m**3")

    p0.render()
    p1.render()

    im0 = p0.plots[field].image._A
    im1 = p1.plots[field].image._A

    npt.assert_allclose(im0, im1)


@pytest.mark.mpl_image_compare
def test_inf_and_finite_values_set_zlim():
    # see https://github.com/yt-project/yt/issues/3901
    shape = (32, 16, 1)
    a = np.ones(16)
    b = np.ones((32, 16))
    c = np.reshape(a * b, shape)

    # injecting an inf
    c[0, 0, 0] = np.inf

    field = ("gas", "density")
    data = {field: c}

    ds = load_uniform_grid(
        data,
        shape,
        bbox=np.array([[0, 1], [0, 1], [0, 1]]),
    )

    p = SlicePlot(ds, "z", field)

    # setting zlim manually
    p.set_zlim(field, -10, 10)

    p.render()
    return p.plots[field].figure


@pytest.mark.skipif(
    MPL_VERSION < Version("3.5"),
    reason=f"Correct behaviour requires MPL 3.5 we have {MPL_VERSION}",
)
@pytest.mark.mpl_image_compare
def test_sliceplot_custom_norm():
    from matplotlib.colors import TwoSlopeNorm

    ds = fake_random_ds(16)
    field = ("gas", "density")
    p = SlicePlot(ds, "z", field)
    p.set_norm(field, norm=TwoSlopeNorm(vcenter=0, vmin=-0.5, vmax=1))

    p.render()
    return p.plots[field].figure


@pytest.mark.skipif(
    MPL_VERSION >= Version("3.5"),
    reason=f"Testing a warning that should only happen with MPL < 3.5, we have {MPL_VERSION}",
)
@pytest.mark.mpl_image_compare
def test_sliceplot_custom_norm_warning():
    from matplotlib.colors import TwoSlopeNorm

    ds = fake_random_ds(16)
    field = ("gas", "density")
    p = SlicePlot(ds, "z", field)
    p.set_norm(field, norm=TwoSlopeNorm(vcenter=0, vmin=-0.5, vmax=1))

    with pytest.warns(
        UserWarning,
        match=re.escape(
            "Failed to format colorbar ticks. "
            "This is expected when using the set_norm method "
            "with some matplotlib classes (e.g. TwoSlopeNorm) "
            "with matplotlib versions older than 3.5\n"
            "Please try upgrading matplotlib to a more recent version. "
            "If the problem persists, please file a report to "
            "https://github.com/yt-project/yt/issues/new"
        ),
    ):
        p.render()
    return p.plots[field].figure


@pytest.mark.mpl_image_compare
def test_lineplot_set_axis_properties():
    ds = fake_random_ds(16)
    field = ("gas", "density")
    p = LinePlot(
        ds,
        field,
        start_point=[0, 0, 0],
        end_point=[1, 1, 1],
        npoints=32,
    )
    p.set_x_unit("cm")
    p.set_unit(field, "kg/cm**3")
    p.set_log(field, False)

    p.render()
    return p.plots[field].figure


@pytest.mark.mpl_image_compare
def test_profileplot_set_axis_properties():
    ds = fake_random_ds(16)

    disk = ds.disk(ds.domain_center, [0.0, 0.0, 1.0], (10, "m"), (1, "m"))
    p = ProfilePlot(disk, ("gas", "density"), [("gas", "velocity_x")])
    p.set_unit(("gas", "density"), "kg/cm**3")
    p.set_log(("gas", "density"), False)
    p.set_unit(("gas", "velocity_x"), "mile/hour")

    p.render()
    return p.plots["gas", "velocity_x"].figure


@pytest.mark.mpl_image_compare
def test_particleprojectionplot_set_colorbar_properties():
    ds = fake_particle_ds(npart=100)

    field = ("all", "particle_mass")
    p = ParticleProjectionPlot(ds, 2, field)
    p.set_buff_size(10)

    p.set_unit(field, "Msun")
    p.set_zlim(field, zmax=1e-35)
    p.set_log(field, False)

    p.render()
    return p.plots[field].figure


class TestProfilePlot:
    @classmethod
    def setup_class(cls):
        fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
        units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
        cls.ds = fake_random_ds(16, fields=fields, units=units)
        regions = [cls.ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), cls.ds.all_data()]
        pr_fields = [
            [("gas", "density"), ("gas", "temperature")],
            [("gas", "density"), ("gas", "velocity_x")],
            [("gas", "temperature"), ("gas", "mass")],
            [("gas", "density"), ("index", "radius")],
            [("gas", "velocity_magnitude"), ("gas", "mass")],
        ]
        cls.profiles: Dict[str, ProfilePlot] = {}
        for i_reg, reg in enumerate(regions):
            id_prefix = str(i_reg)
            for x_field, y_field in pr_fields:
                id_suffix = "_".join([*x_field, *y_field])
                base_id = f"{id_prefix}_{id_suffix}"
                cls.profiles[base_id] = ProfilePlot(reg, x_field, y_field)
                cls.profiles[f"{base_id}_fractional_accumulation"] = ProfilePlot(
                    reg, x_field, y_field, fractional=True, accumulation=True
                )

                p1d = create_profile(reg, x_field, y_field)
                cls.profiles[f"{base_id}_from_profiles"] = ProfilePlot.from_profiles(
                    p1d
                )

        p1 = create_profile(
            cls.ds.all_data(), ("gas", "density"), ("gas", "temperature")
        )
        p2 = create_profile(
            cls.ds.all_data(), ("gas", "density"), ("gas", "velocity_x")
        )
        cls.profiles["from_multiple_profiles"] = ProfilePlot.from_profiles(
            [p1, p2], labels=["temperature", "velocity"]
        )

    @pytest.mark.parametrize(
        "suffix",
        [None, "from_profiles", "fractional_accumulation"],
    )
    @pytest.mark.parametrize("region", ["0", "1"])
    @pytest.mark.parametrize(
        "xax, yax",
        [
            (("gas", "density"), ("gas", "temperature")),
            (("gas", "density"), ("gas", "velocity_x")),
            (("gas", "temperature"), ("gas", "mass")),
            (("gas", "density"), ("index", "radius")),
            (("gas", "velocity_magnitude"), ("gas", "mass")),
        ],
    )
    @pytest.mark.mpl_image_compare
    def test_profileplot_simple(self, region, xax, yax, suffix):
        key = "_".join(chain(region, xax, yax))
        if suffix is not None:
            key += f"_{suffix}"
        plots = list(self.profiles[key].plots.values())
        assert len(plots) == 1
        return plots[0].figure

    @pytest.mark.mpl_image_compare
    def test_profileplot_from_multiple_profiles_0(self):
        plots = list(self.profiles["from_multiple_profiles"].plots.values())
        assert len(plots) == 2
        return plots[0].figure

    @pytest.mark.mpl_image_compare
    def test_profileplot_from_multiple_profiles_1(self):
        plots = list(self.profiles["from_multiple_profiles"].plots.values())
        assert len(plots) == 2
        return plots[1].figure


class TestPhasePlot:
    @classmethod
    def setup_class(cls):
        fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
        units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
        cls.ds = fake_random_ds(16, fields=fields, units=units)
        regions = [cls.ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3), cls.ds.all_data()]
        pr_fields = [
            [("gas", "density"), ("gas", "temperature"), ("gas", "mass")],
            [("gas", "density"), ("gas", "velocity_x"), ("gas", "mass")],
            [
                ("index", "radius"),
                ("gas", "temperature"),
                ("gas", "velocity_magnitude"),
            ],
        ]
        cls.profiles: Dict[str, PhasePlot] = {}
        for i_reg, reg in enumerate(regions):
            id_prefix = str(i_reg)
            for x_field, y_field, z_field in pr_fields:
                id_suffix = "_".join([*x_field, *y_field, *z_field])
                base_id = f"{id_prefix}_{id_suffix}"
                cls.profiles[base_id] = PhasePlot(
                    reg, x_field, y_field, z_field, x_bins=16, y_bins=16
                )
                cls.profiles[f"{base_id}_fractional_accumulation"] = PhasePlot(
                    reg,
                    x_field,
                    y_field,
                    z_field,
                    fractional=True,
                    accumulation=True,
                    x_bins=16,
                    y_bins=16,
                )

                p2d = create_profile(reg, [x_field, y_field], z_field, n_bins=[16, 16])
                cls.profiles[f"{base_id}_from_profiles"] = PhasePlot.from_profile(p2d)

    @pytest.mark.parametrize(
        "suffix",
        [None, "from_profiles", "fractional_accumulation"],
    )
    @pytest.mark.parametrize("region", ["0", "1"])
    @pytest.mark.parametrize(
        "xax, yax, zax",
        [
            (("gas", "density"), ("gas", "temperature"), ("gas", "mass")),
            (("gas", "density"), ("gas", "velocity_x"), ("gas", "mass")),
            (
                ("index", "radius"),
                ("gas", "temperature"),
                ("gas", "velocity_magnitude"),
            ),
        ],
    )
    @pytest.mark.mpl_image_compare
    def test_phaseplot(self, region, xax, yax, zax, suffix):
        key = "_".join(chain(region, xax, yax, zax))
        if suffix is not None:
            key += f"_{suffix}"
        plots = list(self.profiles[key].plots.values())
        assert len(plots) == 1
        return plots[0].figure


class TestPhasePlotSetZlim:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_random_ds(16)
        add_noise_fields(cls.ds)
        cls.data = cls.ds.sphere("c", 1)

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_zlim_with_implicit_units(self):
        p = PhasePlot(
            self.data,
            ("gas", "noise1"),
            ("gas", "noise3"),
            [("gas", "density")],
            weight_field=None,
        )
        field = ("gas", "density")
        p.set_zlim(field, zmax=10)
        p.render()
        return p.plots[field].figure

    @pytest.mark.mpl_image_compare
    def test_phaseplot_set_zlim_with_explicit_units(self):
        p = PhasePlot(
            self.data,
            ("gas", "noise1"),
            ("gas", "noise3"),
            [("gas", "density")],
            weight_field=None,
        )
        field = ("gas", "density")
        # using explicit units, we expect the colorbar units to stay unchanged
        p.set_zlim(field, zmin=(1e36, "kg/AU**3"))
        p.render()
        return p.plots[field].figure


class TestSetBackgroundColor:
    # see https://github.com/yt-project/yt/issues/3854

    @classmethod
    def setup_class(cls):
        cls.ds = fake_random_ds(16)

    @pytest.mark.mpl_image_compare
    def test_sliceplot_set_background_color_linear(self):
        field = ("gas", "density")
        p = SlicePlot(self.ds, "z", field, width=1.5)
        p.set_background_color(field, color="C0")
        p.set_log(field, False)

        p.render()
        return p.plots[field].figure

    @pytest.mark.mpl_image_compare
    def test_sliceplot_set_background_color_log(self):
        field = ("gas", "density")
        p = SlicePlot(self.ds, "z", field, width=1.5)
        p.set_background_color(field, color="C0")

        p.render()
        return p.plots[field].figure


class TestCylindricalZSlicePlot:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_amr_ds(geometry="cylindrical")
        add_noise_fields(cls.ds)
        fields = ["noise%d" % i for i in range(4)]
        cls.plot = SlicePlot(cls.ds, "z", fields)

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_cylindrical_z_log(self, field):
        return self.plot.plots[field].figure

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_cylindrical_z_linear(self, field):
        self.plot.set_log("noise0", False)
        return self.plot.plots[field].figure


class TestSphericalPhiSlicePlot:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_amr_ds(geometry="spherical")
        add_noise_fields(cls.ds)
        fields = ["noise%d" % i for i in range(4)]
        cls.plot = SlicePlot(cls.ds, "phi", fields)

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_spherical_phi_log(self, field):
        return self.plot.plots[field].figure


class TestSphericalThetaSlicePlot:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_amr_ds(geometry="spherical")
        add_noise_fields(cls.ds)
        fields = ["noise%d" % i for i in range(4)]
        cls.plot = SlicePlot(cls.ds, "theta", fields)

    @pytest.mark.parametrize("field", ["noise0", "noise1", "noise2", "noise3"])
    @pytest.mark.mpl_image_compare
    def test_spherical_theta_log(self, field):
        return self.plot.plots[field].figure



# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
# distutils: sources = FIXED_INTERP
# distutils: language = c++
# distutils: extra_compile_args = CPP14_FLAG
# distutils: extra_link_args = CPP14_FLAG
"""
Simple integrators for the radiative transfer equation



"""


import numpy as np

cimport cython
cimport numpy as np
from libc.math cimport atan2, cos, fabs, floor, sin, sqrt

from yt.utilities.lib.fp_utils cimport fmin

from .fixed_interpolator cimport *


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef int walk_volume(VolumeContainer *vc,
                     np.float64_t v_pos[3],
                     np.float64_t v_dir[3],
                     sampler_function *sample,
                     void *data,
                     np.float64_t *return_t = NULL,
                     np.float64_t max_t = 1.0) noexcept nogil:
    cdef int cur_ind[3]
    cdef int step[3]
    cdef int x, y, i, hit, direction
    cdef np.float64_t intersect_t = 1.1
    cdef np.float64_t iv_dir[3]
    cdef np.float64_t tmax[3]
    cdef np.float64_t tdelta[3]
    cdef np.float64_t exit_t = -1.0, enter_t = -1.0
    cdef np.float64_t tl, temp_x, temp_y = -1
    if max_t > 1.0: max_t = 1.0
    direction = -1
    if vc.left_edge[0] <= v_pos[0] and v_pos[0] < vc.right_edge[0] and \
       vc.left_edge[1] <= v_pos[1] and v_pos[1] < vc.right_edge[1] and \
       vc.left_edge[2] <= v_pos[2] and v_pos[2] < vc.right_edge[2]:
        intersect_t = 0.0
        direction = 3
    for i in range(3):
        if (v_dir[i] < 0):
            step[i] = -1
        elif (v_dir[i] == 0.0):
            step[i] = 0
            continue
        else:
            step[i] = 1
        iv_dir[i] = 1.0/v_dir[i]
        if direction == 3: continue
        x = (i+1) % 3
        y = (i+2) % 3
        if step[i] > 0:
            tl = (vc.left_edge[i] - v_pos[i])*iv_dir[i]
        else:
            tl = (vc.right_edge[i] - v_pos[i])*iv_dir[i]
        temp_x = (v_pos[x] + tl*v_dir[x])
        temp_y = (v_pos[y] + tl*v_dir[y])
        if fabs(temp_x - vc.left_edge[x]) < 1e-10*vc.dds[x]:
            temp_x = vc.left_edge[x]
        elif fabs(temp_x - vc.right_edge[x]) < 1e-10*vc.dds[x]:
            temp_x = vc.right_edge[x]
        if fabs(temp_y - vc.left_edge[y]) < 1e-10*vc.dds[y]:
            temp_y = vc.left_edge[y]
        elif fabs(temp_y - vc.right_edge[y]) < 1e-10*vc.dds[y]:
            temp_y = vc.right_edge[y]
        if vc.left_edge[x] <= temp_x and temp_x <= vc.right_edge[x] and \
           vc.left_edge[y] <= temp_y and temp_y <= vc.right_edge[y] and \
           0.0 <= tl and tl < intersect_t:
            direction = i
            intersect_t = tl
    if enter_t >= 0.0: intersect_t = enter_t
    if not ((0.0 <= intersect_t) and (intersect_t < max_t)): return 0
    for i in range(3):
        # Two things have to be set inside this loop.
        # cur_ind[i], the current index of the grid cell the ray is in
        # tmax[i], the 't' until it crosses out of the grid cell
        tdelta[i] = step[i] * iv_dir[i] * vc.dds[i]
        if i == direction and step[i] > 0:
            # Intersection with the left face in this direction
            cur_ind[i] = 0
        elif i == direction and step[i] < 0:
            # Intersection with the right face in this direction
            cur_ind[i] = vc.dims[i] - 1
        else:
            # We are somewhere in the middle of the face
            temp_x = intersect_t * v_dir[i] + v_pos[i] # current position
            temp_y = ((temp_x - vc.left_edge[i])*vc.idds[i])
            # There are some really tough cases where we just within a couple
            # least significant places of the edge, and this helps prevent
            # killing the calculation through a segfault in those cases.
            if -1 < temp_y < 0 and step[i] > 0:
                temp_y = 0.0
            elif vc.dims[i] - 1 < temp_y < vc.dims[i] and step[i] < 0:
                temp_y = vc.dims[i] - 1
            cur_ind[i] =  <int> (floor(temp_y))
        if step[i] > 0:
            temp_y = (cur_ind[i] + 1) * vc.dds[i] + vc.left_edge[i]
        elif step[i] < 0:
            temp_y = cur_ind[i] * vc.dds[i] + vc.left_edge[i]
        tmax[i] = (temp_y - v_pos[i]) * iv_dir[i]
        if step[i] == 0:
            tmax[i] = 1e60
    # We have to jumpstart our calculation
    for i in range(3):
        if cur_ind[i] == vc.dims[i] and step[i] >= 0:
            return 0
        if cur_ind[i] == -1 and step[i] <= -1:
            return 0
    enter_t = intersect_t
    hit = 0
    while 1:
        hit += 1
        if tmax[0] < tmax[1]:
            if tmax[0] < tmax[2]:
                i = 0
            else:
                i = 2
        else:
            if tmax[1] < tmax[2]:
                i = 1
            else:
                i = 2
        exit_t = fmin(tmax[i], max_t)
        sample(vc, v_pos, v_dir, enter_t, exit_t, cur_ind, data)
        cur_ind[i] += step[i]
        enter_t = tmax[i]
        tmax[i] += tdelta[i]
        if cur_ind[i] < 0 or cur_ind[i] >= vc.dims[i] or enter_t >= max_t:
            break
    if return_t != NULL: return_t[0] = exit_t
    return hit

def hp_pix2vec_nest(long nside, long ipix):
    raise NotImplementedError
    # cdef double v[3]
    # healpix_interface.pix2vec_nest(nside, ipix, v)
    # cdef np.ndarray[np.float64_t, ndim=1] tr = np.empty((3,), dtype='float64')
    # tr[0] = v[0]
    # tr[1] = v[1]
    # tr[2] = v[2]
    # return tr

def arr_pix2vec_nest(long nside,
                     np.ndarray[np.int64_t, ndim=1] aipix):
    raise NotImplementedError
    # cdef int n = aipix.shape[0]
    # cdef int i
    # cdef double v[3]
    # cdef long ipix
    # cdef np.ndarray[np.float64_t, ndim=2] tr = np.zeros((n, 3), dtype='float64')
    # for i in range(n):
    #     ipix = aipix[i]
    #     healpix_interface.pix2vec_nest(nside, ipix, v)
    #     tr[i,0] = v[0]
    #     tr[i,1] = v[1]
    #     tr[i,2] = v[2]
    # return tr

def hp_vec2pix_nest(long nside, double x, double y, double z):
    raise NotImplementedError
    # cdef double v[3]
    # v[0] = x
    # v[1] = y
    # v[2] = z
    # cdef long ipix
    # healpix_interface.vec2pix_nest(nside, v, &ipix)
    # return ipix

def arr_vec2pix_nest(long nside,
                     np.ndarray[np.float64_t, ndim=1] x,
                     np.ndarray[np.float64_t, ndim=1] y,
                     np.ndarray[np.float64_t, ndim=1] z):
    raise NotImplementedError
    # cdef int n = x.shape[0]
    # cdef int i
    # cdef double v[3]
    # cdef long ipix
    # cdef np.ndarray[np.int64_t, ndim=1] tr = np.zeros(n, dtype='int64')
    # for i in range(n):
    #     v[0] = x[i]
    #     v[1] = y[i]
    #     v[2] = z[i]
    #     healpix_interface.vec2pix_nest(nside, v, &ipix)
    #     tr[i] = ipix
    # return tr

def hp_pix2ang_nest(long nside, long ipnest):
    raise NotImplementedError
    # cdef double theta, phi
    # healpix_interface.pix2ang_nest(nside, ipnest, &theta, &phi)
    # return (theta, phi)

def arr_pix2ang_nest(long nside, np.ndarray[np.int64_t, ndim=1] aipnest):
    raise NotImplementedError
    # cdef int n = aipnest.shape[0]
    # cdef int i
    # cdef long ipnest
    # cdef np.ndarray[np.float64_t, ndim=2] tr = np.zeros((n, 2), dtype='float64')
    # cdef double theta, phi
    # for i in range(n):
    #     ipnest = aipnest[i]
    #     healpix_interface.pix2ang_nest(nside, ipnest, &theta, &phi)
    #     tr[i,0] = theta
    #     tr[i,1] = phi
    # return tr

def hp_ang2pix_nest(long nside, double theta, double phi):
    raise NotImplementedError
    # cdef long ipix
    # healpix_interface.ang2pix_nest(nside, theta, phi, &ipix)
    # return ipix

def arr_ang2pix_nest(long nside,
                     np.ndarray[np.float64_t, ndim=1] atheta,
                     np.ndarray[np.float64_t, ndim=1] aphi):
    raise NotImplementedError
    # cdef int n = atheta.shape[0]
    # cdef int i
    # cdef long ipnest
    # cdef np.ndarray[np.int64_t, ndim=1] tr = np.zeros(n, dtype='int64')
    # cdef double theta, phi
    # for i in range(n):
    #     theta = atheta[i]
    #     phi = aphi[i]
    #     healpix_interface.ang2pix_nest(nside, theta, phi, &ipnest)
    #     tr[i] = ipnest
    # return tr

@cython.boundscheck(False)
@cython.cdivision(False)
@cython.wraparound(False)
def pixelize_healpix(long nside,
                     np.ndarray[np.float64_t, ndim=1] values,
                     long ntheta, long nphi,
                     np.ndarray[np.float64_t, ndim=2] irotation):
    raise NotImplementedError
    # # We will first to pix2vec, rotate, then calculate the angle
    # cdef int i, j, thetai, phii
    # cdef long ipix
    # cdef double v0[3], v1[3]
    # cdef double pi = 3.1415926
    # cdef np.float64_t pi2 = pi/2.0
    # cdef np.float64_t phi, theta
    # cdef np.ndarray[np.float64_t, ndim=2] results
    # cdef np.ndarray[np.int32_t, ndim=2] count
    # results = np.zeros((ntheta, nphi), dtype="float64")
    # count = np.zeros((ntheta, nphi), dtype="int32")

    # cdef np.float64_t phi0 = 0
    # cdef np.float64_t dphi = 2.0 * pi/(nphi-1)

    # cdef np.float64_t theta0 = 0
    # cdef np.float64_t dtheta = pi/(ntheta-1)
    # # We assume these are the rotated theta and phi
    # for thetai in range(ntheta):
    #     theta = theta0 + dtheta * thetai
    #     for phii in range(nphi):
    #         phi = phi0 + dphi * phii
    #         # We have our rotated vector
    #         v1[0] = cos(phi) * sin(theta)
    #         v1[1] = sin(phi) * sin(theta)
    #         v1[2] = cos(theta)
    #         # Now we rotate back
    #         for i in range(3):
    #             v0[i] = 0
    #             for j in range(3):
    #                 v0[i] += v1[j] * irotation[j,i]
    #         # Get the pixel this vector is inside
    #         healpix_interface.vec2pix_nest(nside, v0, &ipix)
    #         results[thetai, phii] = values[ipix]
    #         count[i, j] += 1
    # return results, count

def healpix_aitoff_proj(np.ndarray[np.float64_t, ndim=1] pix_image,
                        long nside,
                        np.ndarray[np.float64_t, ndim=2] image,
                        np.ndarray[np.float64_t, ndim=2] irotation):
    raise NotImplementedError
    # cdef double pi = np.pi
    # cdef int i, j, k, l
    # cdef np.float64_t x, y, z, zb
    # cdef np.float64_t dx, dy, inside
    # cdef double v0[3], v1[3]
    # dx = 2.0 / (image.shape[1] - 1)
    # dy = 2.0 / (image.shape[0] - 1)
    # cdef np.float64_t s2 = sqrt(2.0)
    # cdef long ipix
    # for i in range(image.shape[1]):
    #     x = (-1.0 + i*dx)*s2*2.0
    #     for j in range(image.shape[0]):
    #         y = (-1.0 + j * dy)*s2
    #         zb = (x*x/8.0 + y*y/2.0 - 1.0)
    #         if zb > 0: continue
    #         z = (1.0 - (x/4.0)**2.0 - (y/2.0)**2.0)
    #         z = sqrt(z)
    #         # Longitude
    #         phi = (2.0*atan(z*x/(2.0 * (2.0*z*z-1.0))) + pi)
    #         # Latitude
    #         # We shift it into co-latitude
    #         theta = (asin(z*y) + pi/2.0)
    #         # Now to account for rotation we translate into vectors
    #         v1[0] = cos(phi) * sin(theta)
    #         v1[1] = sin(phi) * sin(theta)
    #         v1[2] = cos(theta)
    #         for k in range(3):
    #             v0[k] = 0
    #             for l in range(3):
    #                 v0[k] += v1[l] * irotation[l,k]
    #         healpix_interface.vec2pix_nest(nside, v0, &ipix)
    #         #print("Rotated", v0[0], v0[1], v0[2], v1[0], v1[1], v1[2], ipix, pix_image[ipix])
    #         image[j, i] = pix_image[ipix]

def arr_fisheye_vectors(int resolution, np.float64_t fov, int nimx=1, int
        nimy=1, int nimi=0, int nimj=0, np.float64_t off_theta=0.0, np.float64_t
        off_phi=0.0):
    # We now follow figures 4-7 of:
    # http://paulbourke.net/miscellaneous/domefisheye/fisheye/
    # ...but all in Cython.
    cdef np.ndarray[np.float64_t, ndim=3] vp
    cdef int i, j
    cdef np.float64_t r, phi, theta, px, py
    cdef np.float64_t fov_rad = fov * np.pi / 180.0
    cdef int nx = resolution//nimx
    cdef int ny = resolution//nimy
    vp = np.zeros((nx,ny, 3), dtype="float64")
    for i in range(nx):
        px = (2.0 * (nimi*nx + i)) / resolution - 1.0
        for j in range(ny):
            py = (2.0 * (nimj*ny + j)) / resolution - 1.0
            r = sqrt(px*px + py*py)
            if r > 1.01:
                vp[i,j,0] = vp[i,j,1] = vp[i,j,2] = 0.0
                continue
            phi = atan2(py, px)
            theta = r * fov_rad / 2.0
            theta += off_theta
            phi += off_phi
            vp[i,j,0] = sin(theta) * cos(phi)
            vp[i,j,1] = sin(theta) * sin(phi)
            vp[i,j,2] = cos(theta)
    return vp



import os
import shutil
import tempfile
import unittest

from numpy.testing import assert_raises

import yt
from yt.config import ytcfg
from yt.testing import fake_random_ds
from yt.utilities.configure import YTConfig, config_dir

_TEST_PLUGIN = "_test_plugin.py"

_DUMMY_CFG_TOML = f"""[yt]
log_level = 49
plugin_filename = "{_TEST_PLUGIN}"
boolean_stuff = true
chunk_size = 3
"""

TEST_PLUGIN_FILE = """
import numpy as np

def _myfunc(field, data):
    return np.random.random(data[('gas', 'density')].shape)
add_field(('gas', 'random'), dimensions='dimensionless',
          function=_myfunc, units='auto', sampling_type='local')
constant = 3
def myfunc():
    return constant*4
foobar = 17
"""


class TestPluginFile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.xdg_config_home = os.environ.get("XDG_CONFIG_HOME")
        cls.tmpdir = tempfile.mkdtemp()
        os.mkdir(os.path.join(cls.tmpdir, "yt"))
        os.environ["XDG_CONFIG_HOME"] = cls.tmpdir
        with open(YTConfig.get_global_config_file(), mode="w") as fh:
            fh.write(_DUMMY_CFG_TOML)
        cls.plugin_path = os.path.join(config_dir(), ytcfg.get("yt", "plugin_filename"))
        with open(cls.plugin_path, mode="w") as fh:
            fh.write(TEST_PLUGIN_FILE)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)
        if cls.xdg_config_home:
            os.environ["XDG_CONFIG_HOME"] = cls.xdg_config_home
        else:
            os.environ.pop("XDG_CONFIG_HOME")

    def testCustomField(self):
        msg = f"INFO:yt:Loading plugins from {self.plugin_path}"
        with self.assertLogs("yt", level="INFO") as cm:
            yt.enable_plugins()
            self.assertEqual(cm.output, [msg])

        ds = fake_random_ds(16)
        dd = ds.all_data()
        self.assertEqual(str(dd[("gas", "random")].units), "dimensionless")
        self.assertEqual(dd[("gas", "random")].shape, dd[("gas", "density")].shape)
        assert yt.myfunc() == 12
        assert_raises(AttributeError, getattr, yt, "foobar")



import numpy as np

from yt._typing import KnownFieldsT
from yt.fields.field_info_container import (
    FieldInfoContainer,
    particle_deposition_functions,
    particle_vector_functions,
    standard_particle_fields,
)
from yt.units.unit_object import Unit  # type: ignore
from yt.utilities.exceptions import YTFieldNotFound

rho_units = "code_mass / code_length**3"
mom_units = "code_mass / (code_time * code_length**2)"
eden_units = "code_mass / (code_time**2 * code_length)"  # erg / cm^3
vel_units = "code_length / code_time"
b_units = "code_magnetic"


class ChomboFieldInfo(FieldInfoContainer):
    # no custom behaviour is needed yet
    pass


# Orion 2 Fields
# We duplicate everything here from Boxlib, because we want to be able to
# subclass it and that can be somewhat tricky.
class Orion2FieldInfo(ChomboFieldInfo):
    known_other_fields: KnownFieldsT = (
        ("density", (rho_units, ["density"], None)),
        ("energy-density", (eden_units, ["total_energy_density"], None)),
        ("radiation-energy-density", (eden_units, ["radiation_energy_density"], None)),
        ("X-momentum", (mom_units, ["momentum_density_x"], None)),
        ("Y-momentum", (mom_units, ["momentum_density_y"], None)),
        ("Z-momentum", (mom_units, ["momentum_density_z"], None)),
        ("temperature", ("K", ["temperature"], None)),
        ("X-magnfield", (b_units, [], None)),
        ("Y-magnfield", (b_units, [], None)),
        ("Z-magnfield", (b_units, [], None)),
        ("directrad-dedt-density", (eden_units, ["directrad-dedt-density"], None)),
        ("directrad-dpxdt-density", (mom_units, ["directrad-dpxdt-density"], None)),
        ("directrad-dpydt-density", (mom_units, ["directrad-dpydt-density"], None)),
        ("directrad-dpzdt-density", (mom_units, ["directrad-dpzdt-density"], None)),
    )
    known_particle_fields: KnownFieldsT = (
        ("particle_mass", ("code_mass", [], None)),
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_momentum_x", ("code_mass*code_length/code_time", [], None)),
        ("particle_momentum_y", ("code_mass*code_length/code_time", [], None)),
        ("particle_momentum_z", ("code_mass*code_length/code_time", [], None)),
        # Note that these are *internal* agmomen
        ("particle_angmomen_x", ("code_length**2/code_time", [], None)),
        ("particle_angmomen_y", ("code_length**2/code_time", [], None)),
        ("particle_angmomen_z", ("code_length**2/code_time", [], None)),
        ("particle_mlast", ("code_mass", [], None)),
        ("particle_r", ("code_length", [], None)),
        ("particle_mdeut", ("code_mass", [], None)),
        ("particle_n", ("", [], None)),
        ("particle_mdot", ("code_mass/code_time", [], None)),
        ("particle_burnstate", ("", [], None)),
        ("particle_luminosity", ("", [], None)),
        ("particle_id", ("", ["particle_index"], None)),
    )

    def setup_particle_fields(self, ptype):
        def _get_vel(axis):
            def velocity(field, data):
                return (
                    data[(ptype, f"particle_momentum_{axis}")]
                    / data[(ptype, "particle_mass")]
                )

            return velocity

        for ax in "xyz":
            self.add_field(
                (ptype, f"particle_velocity_{ax}"),
                sampling_type="particle",
                function=_get_vel(ax),
                units="code_length/code_time",
            )

        super().setup_particle_fields(ptype)

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        unit_system = self.ds.unit_system

        def _thermal_energy_density(field, data):
            try:
                return (
                    data[("chombo", "energy-density")]
                    - data[("gas", "kinetic_energy_density")]
                    - data[("gas", "magnetic_energy_density")]
                )
            except YTFieldNotFound:
                return (
                    data[("chombo", "energy-density")]
                    - data[("gas", "kinetic_energy_density")]
                )

        def _specific_thermal_energy(field, data):
            return data[("gas", "thermal_energy_density")] / data[("gas", "density")]

        def _magnetic_energy_density(field, data):
            ret = data[("chombo", "X-magnfield")] ** 2
            if data.ds.dimensionality > 1:
                ret = ret + data[("chombo", "Y-magnfield")] ** 2
            if data.ds.dimensionality > 2:
                ret = ret + data[("chombo", "Z-magnfield")] ** 2
            return ret / 8.0 / np.pi

        def _specific_magnetic_energy(field, data):
            return data[("gas", "specific_magnetic_energy")] / data[("gas", "density")]

        def _kinetic_energy_density(field, data):
            p2 = data[("chombo", "X-momentum")] ** 2
            if data.ds.dimensionality > 1:
                p2 = p2 + data[("chombo", "Y-momentum")] ** 2
            if data.ds.dimensionality > 2:
                p2 = p2 + data[("chombo", "Z-momentum")] ** 2
            return 0.5 * p2 / data[("gas", "density")]

        def _specific_kinetic_energy(field, data):
            return data[("gas", "kinetic_energy_density")] / data[("gas", "density")]

        def _temperature(field, data):
            c_v = data.ds.quan(data.ds.parameters["radiation.const_cv"], "erg/g/K")
            return data[("gas", "specific_thermal_energy")] / c_v

        def _get_vel(axis):
            def velocity(field, data):
                return (
                    data[("gas", f"momentum_density_{axis}")] / data[("gas", "density")]
                )

            return velocity

        for ax in "xyz":
            self.add_field(
                ("gas", f"velocity_{ax}"),
                sampling_type="cell",
                function=_get_vel(ax),
                units=unit_system["velocity"],
            )
        self.add_field(
            ("gas", "specific_thermal_energy"),
            sampling_type="cell",
            function=_specific_thermal_energy,
            units=unit_system["specific_energy"],
        )
        self.add_field(
            ("gas", "thermal_energy_density"),
            sampling_type="cell",
            function=_thermal_energy_density,
            units=unit_system["pressure"],
        )
        self.add_field(
            ("gas", "kinetic_energy_density"),
            sampling_type="cell",
            function=_kinetic_energy_density,
            units=unit_system["pressure"],
        )
        self.add_field(
            ("gas", "specific_kinetic_energy"),
            sampling_type="cell",
            function=_specific_kinetic_energy,
            units=unit_system["specific_energy"],
        )
        self.add_field(
            ("gas", "magnetic_energy_density"),
            sampling_type="cell",
            function=_magnetic_energy_density,
            units=unit_system["pressure"],
        )
        self.add_field(
            ("gas", "specific_magnetic_energy"),
            sampling_type="cell",
            function=_specific_magnetic_energy,
            units=unit_system["specific_energy"],
        )
        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=unit_system["temperature"],
        )

        setup_magnetic_field_aliases(
            self, "chombo", [f"{ax}-magnfield" for ax in "XYZ"]
        )


class ChomboPICFieldInfo3D(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("density", (rho_units, ["density", "Density"], None)),
        (
            "potential",
            ("code_length**2 / code_time**2", ["potential", "Potential"], None),
        ),
        ("gravitational_field_x", ("code_length / code_time**2", [], None)),
        ("gravitational_field_y", ("code_length / code_time**2", [], None)),
        ("gravitational_field_z", ("code_length / code_time**2", [], None)),
    )
    known_particle_fields: KnownFieldsT = (
        ("particle_mass", ("code_mass", [], None)),
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", ("code_length / code_time", [], None)),
        ("particle_velocity_y", ("code_length / code_time", [], None)),
        ("particle_velocity_z", ("code_length / code_time", [], None)),
    )

    # I am re-implementing this here to override a few default behaviors:
    # I don't want to skip output units for code_length and I do want
    # particle_fields to default to take_log = False.
    def setup_particle_fields(self, ptype, ftype="gas", num_neighbors=64):
        skip_output_units = ()
        for f, (units, aliases, dn) in sorted(self.known_particle_fields):
            units = self.ds.field_units.get((ptype, f), units)
            if (
                f in aliases or ptype not in self.ds.particle_types_raw
            ) and units not in skip_output_units:
                u = Unit(units, registry=self.ds.unit_registry)
                output_units = str(u.get_cgs_equivalent())
            else:
                output_units = units
            if (ptype, f) not in self.field_list:
                continue
            self.add_output_field(
                (ptype, f),
                sampling_type="particle",
                units=units,
                display_name=dn,
                output_units=output_units,
                take_log=False,
            )
            for alias in aliases:
                self.alias((ptype, alias), (ptype, f), units=output_units)

        ppos_fields = [f"particle_position_{ax}" for ax in "xyz"]
        pvel_fields = [f"particle_velocity_{ax}" for ax in "xyz"]
        particle_vector_functions(ptype, ppos_fields, pvel_fields, self)

        particle_deposition_functions(ptype, "particle_position", "particle_mass", self)
        standard_particle_fields(self, ptype)
        # Now we check for any leftover particle fields
        for field in sorted(self.field_list):
            if field in self:
                continue
            if not isinstance(field, tuple):
                raise RuntimeError
            if field[0] not in self.ds.particle_types:
                continue
            self.add_output_field(
                field,
                sampling_type="particle",
                units=self.ds.field_units.get(field, ""),
            )
        self.setup_smoothed_fields(ptype, num_neighbors=num_neighbors, ftype=ftype)


def _dummy_position(field, data):
    return 0.5 * np.ones_like(data[("all", "particle_position_x")])


def _dummy_velocity(field, data):
    return np.zeros_like(data[("all", "particle_velocity_x")])


def _dummy_field(field, data):
    return 0.0 * data[("chombo", "gravitational_field_x")]


fluid_field_types = ["chombo", "gas"]
particle_field_types = ["io", "all"]


class ChomboPICFieldInfo2D(ChomboPICFieldInfo3D):
    known_other_fields: KnownFieldsT = (
        ("density", (rho_units, ["density", "Density"], None)),
        (
            "potential",
            ("code_length**2 / code_time**2", ["potential", "Potential"], None),
        ),
        ("gravitational_field_x", ("code_length / code_time**2", [], None)),
        ("gravitational_field_y", ("code_length / code_time**2", [], None)),
    )
    known_particle_fields: KnownFieldsT = (
        ("particle_mass", ("code_mass", [], None)),
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_velocity_x", ("code_length / code_time", [], None)),
        ("particle_velocity_y", ("code_length / code_time", [], None)),
    )

    def __init__(self, ds, field_list):
        super().__init__(ds, field_list)

        for ftype in fluid_field_types:
            self.add_field(
                (ftype, "gravitational_field_z"),
                sampling_type="cell",
                function=_dummy_field,
                units="code_length / code_time**2",
            )

        for ptype in particle_field_types:
            self.add_field(
                (ptype, "particle_position_z"),
                sampling_type="particle",
                function=_dummy_position,
                units="code_length",
            )

            self.add_field(
                (ptype, "particle_velocity_z"),
                sampling_type="particle",
                function=_dummy_velocity,
                units="code_length / code_time",
            )


class ChomboPICFieldInfo1D(ChomboPICFieldInfo3D):
    known_other_fields: KnownFieldsT = (
        ("density", (rho_units, ["density", "Density"], None)),
        (
            "potential",
            ("code_length**2 / code_time**2", ["potential", "Potential"], None),
        ),
        ("gravitational_field_x", ("code_length / code_time**2", [], None)),
    )
    known_particle_fields: KnownFieldsT = (
        ("particle_mass", ("code_mass", [], None)),
        ("particle_position_x", ("code_length", [], None)),
        ("particle_velocity_x", ("code_length / code_time", [], None)),
    )

    def __init__(self, ds, field_list):
        super().__init__(ds, field_list)

        for ftype in fluid_field_types:
            self.add_field(
                (ftype, "gravitational_field_y"),
                sampling_type="cell",
                function=_dummy_field,
                units="code_length / code_time**2",
            )

            self.add_field(
                (ftype, "gravitational_field_z"),
                sampling_type="cell",
                function=_dummy_field,
                units="code_length / code_time**2",
            )

        for ptype in particle_field_types:
            self.add_field(
                (ptype, "particle_position_y"),
                sampling_type="particle",
                function=_dummy_position,
                units="code_length",
            )
            self.add_field(
                (ptype, "particle_position_z"),
                sampling_type="particle",
                function=_dummy_position,
                units="code_length",
            )
            self.add_field(
                (ptype, "particle_velocity_y"),
                sampling_type="particle",
                function=_dummy_velocity,
                units="code_length / code_time",
            )
            self.add_field(
                (ptype, "particle_velocity_z"),
                sampling_type="particle",
                function=_dummy_velocity,
                units="code_length / code_time",
            )


class PlutoFieldInfo(ChomboFieldInfo):
    known_other_fields: KnownFieldsT = (
        ("rho", (rho_units, ["density"], None)),
        ("prs", ("code_mass / (code_length * code_time**2)", ["pressure"], None)),
        ("vx1", (vel_units, ["velocity_x"], None)),
        ("vx2", (vel_units, ["velocity_y"], None)),
        ("vx3", (vel_units, ["velocity_z"], None)),
        ("bx1", (b_units, [], None)),
        ("bx2", (b_units, [], None)),
        ("bx3", (b_units, [], None)),
    )

    known_particle_fields = ()

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        setup_magnetic_field_aliases(self, "chombo", [f"bx{ax}" for ax in [1, 2, 3]])



# distutils: libraries = STD_LIBS
# distutils: language = c++
# distutils: extra_compile_args = CPP14_FLAG OMP_ARGS
# distutils: extra_link_args = CPP14_FLAG OMP_ARGS
"""
Simple utilities that don't fit anywhere else



"""


import numpy as np

from yt.funcs import get_pbar
from yt.units.yt_array import YTArray

cimport cython
cimport numpy as np
from cpython cimport buffer
from cython.view cimport memoryview
from libc.math cimport abs, sqrt
from libc.stdlib cimport free, malloc
from libc.string cimport strcmp

from yt.geometry.selection_routines cimport _ensure_code
from yt.utilities.lib.fp_utils cimport fmax, fmin


cdef extern from "platform_dep.h":
    # NOTE that size_t might not be int
    void *alloca(int)

from cython.parallel import prange

from cpython.exc cimport PyErr_CheckSignals


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def new_bin_profile1d(np.ndarray[np.intp_t, ndim=1] bins_x,
                  np.ndarray[np.float64_t, ndim=1] wsource,
                  np.ndarray[np.float64_t, ndim=2] bsource,
                  np.ndarray[np.float64_t, ndim=1] wresult,
                  np.ndarray[np.float64_t, ndim=2] bresult,
                  np.ndarray[np.float64_t, ndim=2] mresult,
                  np.ndarray[np.float64_t, ndim=2] qresult,
                  np.ndarray[np.uint8_t, ndim=1, cast=True] used):
    cdef int n, fi, bin
    cdef np.float64_t wval, bval, oldwr, bval_mresult
    cdef int nb = bins_x.shape[0]
    cdef int nf = bsource.shape[1]
    for n in range(nb):
        bin = bins_x[n]
        wval = wsource[n]
        # Skip field value entries where the weight field is zero
        if wval == 0:
            continue
        oldwr = wresult[bin]
        wresult[bin] += wval
        for fi in range(nf):
            bval = bsource[n,fi]
            bval_mresult = bval - mresult[bin,fi]
            # qresult has to have the previous wresult
            qresult[bin,fi] += oldwr * wval * bval_mresult * bval_mresult / \
                (oldwr + wval)
            bresult[bin,fi] += wval*bval
            # mresult needs the new wresult
            mresult[bin,fi] += wval * bval_mresult / wresult[bin]
        used[bin] = 1
    return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def new_bin_profile2d(np.ndarray[np.intp_t, ndim=1] bins_x,
                  np.ndarray[np.intp_t, ndim=1] bins_y,
                  np.ndarray[np.float64_t, ndim=1] wsource,
                  np.ndarray[np.float64_t, ndim=2] bsource,
                  np.ndarray[np.float64_t, ndim=2] wresult,
                  np.ndarray[np.float64_t, ndim=3] bresult,
                  np.ndarray[np.float64_t, ndim=3] mresult,
                  np.ndarray[np.float64_t, ndim=3] qresult,
                  np.ndarray[np.uint8_t, ndim=2, cast=True] used):
    cdef int n, fi, bin_x, bin_y
    cdef np.float64_t wval, bval, oldwr, bval_mresult
    cdef int nb = bins_x.shape[0]
    cdef int nf = bsource.shape[1]
    for n in range(nb):
        bin_x = bins_x[n]
        bin_y = bins_y[n]
        wval = wsource[n]
        # Skip field value entries where the weight field is zero
        if wval == 0:
            continue
        oldwr = wresult[bin_x, bin_y]
        wresult[bin_x,bin_y] += wval
        for fi in range(nf):
            bval = bsource[n,fi]
            bval_mresult = bval - mresult[bin_x,bin_y,fi]
            # qresult has to have the previous wresult
            qresult[bin_x,bin_y,fi] += oldwr * wval * bval_mresult * bval_mresult / \
                (oldwr + wval)
            bresult[bin_x,bin_y,fi] += wval*bval
            # mresult needs the new wresult
            mresult[bin_x,bin_y,fi] += wval * bval_mresult / wresult[bin_x,bin_y]
        used[bin_x,bin_y] = 1
    return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def new_bin_profile3d(np.ndarray[np.intp_t, ndim=1] bins_x,
                  np.ndarray[np.intp_t, ndim=1] bins_y,
                  np.ndarray[np.intp_t, ndim=1] bins_z,
                  np.ndarray[np.float64_t, ndim=1] wsource,
                  np.ndarray[np.float64_t, ndim=2] bsource,
                  np.ndarray[np.float64_t, ndim=3] wresult,
                  np.ndarray[np.float64_t, ndim=4] bresult,
                  np.ndarray[np.float64_t, ndim=4] mresult,
                  np.ndarray[np.float64_t, ndim=4] qresult,
                  np.ndarray[np.uint8_t, ndim=3, cast=True] used):
    cdef int n, fi, bin_x, bin_y, bin_z
    cdef np.float64_t wval, bval, oldwr, bval_mresult
    cdef int nb = bins_x.shape[0]
    cdef int nf = bsource.shape[1]
    for n in range(nb):
        bin_x = bins_x[n]
        bin_y = bins_y[n]
        bin_z = bins_z[n]
        wval = wsource[n]
        # Skip field value entries where the weight field is zero
        if wval == 0:
            continue
        oldwr = wresult[bin_x, bin_y, bin_z]
        wresult[bin_x,bin_y,bin_z] += wval
        for fi in range(nf):
            bval = bsource[n,fi]
            bval_mresult = bval - mresult[bin_x,bin_y,bin_z,fi]
            # qresult has to have the previous wresult
            qresult[bin_x,bin_y,bin_z,fi] += \
                oldwr * wval * bval_mresult * bval_mresult / \
                (oldwr + wval)
            bresult[bin_x,bin_y,bin_z,fi] += wval*bval
            # mresult needs the new wresult
            mresult[bin_x,bin_y,bin_z,fi] += wval * bval_mresult / \
                 wresult[bin_x,bin_y,bin_z]
        used[bin_x,bin_y,bin_z] = 1
    return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def lines(np.float64_t[:,:,:] image,
          np.int64_t[:] xs,
          np.int64_t[:] ys,
          np.float64_t[:,:] colors,
          int points_per_color=1,
          int thick=1,
	      int flip=0,
          int crop = 0):

    cdef int nx = image.shape[0]
    cdef int ny = image.shape[1]
    cdef int nl = xs.shape[0]
    cdef np.float64_t alpha[4]
    cdef np.float64_t outa
    cdef int i, j, xi, yi
    cdef int dx, dy, sx, sy, e2, err
    cdef np.int64_t x0, x1, y0, y1
    cdef int has_alpha = (image.shape[2] == 4)
    cdef int no_color = (image.shape[2] < 3)
    for j in range(0, nl, 2):
        # From wikipedia http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
        x0 = xs[j]
        y0 = ys[j]
        x1 = xs[j+1]
        y1 = ys[j+1]
        dx = abs(x1-x0)
        dy = abs(y1-y0)
        if crop == 1 and (dx > nx/2.0 or dy > ny/2.0):
            continue
        err = dx - dy

        if no_color:
            for i in range(4):
                alpha[i] = colors[j, 0]
        elif has_alpha:
            for i in range(4):
                alpha[i] = colors[j/points_per_color,i]
        else:
            for i in range(3):
                alpha[i] = colors[j/points_per_color,3]*\
                        colors[j/points_per_color,i]

        if x0 < x1:
            sx = 1
        else:
            sx = -1
        if y0 < y1:
            sy = 1
        else:
            sy = -1
        while(1):
            if (x0 < thick and sx == -1): break
            elif (x0 >= nx-thick+1 and sx == 1): break
            elif (y0 < thick and sy == -1): break
            elif (y0 >= ny-thick+1 and sy == 1): break
            if x0 >= thick and x0 < nx-thick and y0 >= thick and y0 < ny-thick:
                for xi in range(x0-thick/2, x0+(1+thick)/2):
                    for yi in range(y0-thick/2, y0+(1+thick)/2):
                        if flip:
                            yi0 = ny - yi
                        else:
                            yi0 = yi

                        if no_color:
                            image[xi, yi0, 0] = fmin(alpha[0], image[xi, yi0, 0])
                        elif has_alpha:
                            image[xi, yi0, 3] = outa = alpha[3] + image[xi, yi0, 3]*(1-alpha[3])
                            if outa != 0.0:
                                outa = 1.0/outa
                            for i in range(3):
                                image[xi, yi0, i] = \
                                        ((1.-alpha[3])*image[xi, yi0, i]*image[xi, yi0, 3]
                                         + alpha[3]*alpha[i])*outa
                        else:
                            for i in range(3):
                                image[xi, yi0, i] = \
                                        (1.-alpha[i])*image[xi,yi0,i] + alpha[i]


            if (x0 == x1 and y0 == y1):
                break
            e2 = 2*err
            if e2 > -dy:
                err = err - dy
                x0 += sx
            if e2 < dx :
                err = err + dx
                y0 += sy
    return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def zlines(np.ndarray[np.float64_t, ndim=3] image,
        np.ndarray[np.float64_t, ndim=2] zbuffer,
        np.ndarray[np.int64_t, ndim=1] xs,
        np.ndarray[np.int64_t, ndim=1] ys,
        np.ndarray[np.float64_t, ndim=1] zs,
        np.ndarray[np.float64_t, ndim=2] colors,
        int points_per_color=1,
        int thick=1,
        int flip=0,
        int crop = 0):

    cdef int nx = image.shape[0]
    cdef int ny = image.shape[1]
    cdef int nl = xs.shape[0]
    cdef np.float64_t[:] alpha
    cdef int i, j, c
    cdef int dx, dy, sx, sy, e2, err
    cdef np.int64_t x0, x1, y0, y1, yi0
    cdef np.float64_t z0, z1, dzx, dzy
    alpha = np.zeros(4)
    for j in range(0, nl, 2):
        # From wikipedia http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
        x0 = xs[j]
        y0 = ys[j]
        x1 = xs[j+1]
        y1 = ys[j+1]
        z0 = zs[j]
        z1 = zs[j+1]
        dx = abs(x1-x0)
        dy = abs(y1-y0)
        dzx = (z1-z0) / (dx * dx + dy * dy) * dx
        dzy = (z1-z0) / (dx * dx + dy * dy) * dy
        err = dx - dy
        if crop == 1 and (dx > nx/2.0 or dy > ny/2.0):
            continue

        c = j/points_per_color/2

        for i in range(3):
            alpha[i] = colors[c, i] * colors[c, 3]
        alpha[3] = colors[c, 3]

        if x0 < x1:
            sx = 1
        else:
            sx = -1
        if y0 < y1:
            sy = 1
        else:
            sy = -1
        while(1):
            if (x0 < thick and sx == -1): break
            elif (x0 >= nx-thick+1 and sx == 1): break
            elif (y0 < thick and sy == -1): break
            elif (y0 >= ny-thick+1 and sy == 1): break
            if x0 >= thick and x0 < nx-thick and y0 >= thick and y0 < ny-thick:
                for _ in range(x0-thick/2, x0+(1+thick)/2):
                    for yi in range(y0-thick/2, y0+(1+thick)/2):
                        if flip:
                            yi0 = ny - yi
                        else:
                            yi0 = yi
                        if z0 < zbuffer[x0, yi0]:
                            if alpha[3] != 1.0:
                                talpha = image[x0, yi0, 3]
                                image[x0, yi0, 3] = alpha[3] + talpha * (1 - alpha[3])
                                for i in range(3):
                                    if image[x0, yi0, 3] == 0.0:
                                        image[x0, yi0, i] = 0.0
                                    else:
                                        image[x0, yi0, i] = (alpha[3]*alpha[i] + image[x0, yi0, i]*talpha*(1.0-alpha[3]))/image[x0,yi0,3]
                            else:
                                for i in range(4):
                                    image[x0, yi0, i] = alpha[i]
                            if (1.0 - image[x0, yi0, 3] < 1.0e-4):
                                image[x0, yi0, 3] = 1.0
                                zbuffer[x0, yi0] = z0

            if (x0 == x1 and y0 == y1):
                break
            e2 = 2*err
            if e2 > -dy:
                err = err - dy
                x0 += sx
                z0 += dzx
            if e2 < dx :
                err = err + dx
                y0 += sy
                z0 += dzy
        # assert(np.abs(z0 - z1) < 1.0e-3 * (np.abs(z0) + np.abs(z1)))
    return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def zpoints(np.ndarray[np.float64_t, ndim=3] image,
        np.ndarray[np.float64_t, ndim=2] zbuffer,
        np.ndarray[np.int64_t, ndim=1] xs,
        np.ndarray[np.int64_t, ndim=1] ys,
        np.ndarray[np.float64_t, ndim=1] zs,
        np.ndarray[np.float64_t, ndim=2] colors,
        np.ndarray[np.int64_t, ndim=1] radii, #pixels
        int points_per_color=1,
        int thick=1,
        int flip=0):

    cdef int nx = image.shape[0]
    cdef int ny = image.shape[1]
    cdef np.float64_t[:] alpha
    cdef np.float64_t talpha
    cdef int i, j, c
    cdef np.int64_t kx, ky, r, r2
    cdef np.int64_t[:] idx, ks
    cdef np.int64_t x0, y0, yi0
    cdef np.float64_t z0
    alpha = np.zeros(4)
    #the sources must be ordered along z to avoid edges when two overlap
    idx = np.argsort(zs)
    for j in idx:
        r = radii[j]
        r2 = int((r+0.3)*(r+0.3))  #0.3 to get nicer shape
        ks = np.arange(-r, r+1, dtype=np.int64)
        z0 = zs[j]
        for kx in ks:
            x0 = xs[j]+kx
            if (x0 < 0 or x0 >= nx): continue
            for ky in ks:
                y0 = ys[j]+ky
                if (y0 < 0 or y0 >= ny): continue
                if (kx*kx + ky*ky > r2): continue

                c = j/points_per_color
                for i in range(3):
                    alpha[i] = colors[c, i] * colors[c, 3]
                alpha[3] = colors[c, 3]
                if flip:
                    yi0 = ny - y0
                else:
                    yi0 = y0

                if z0 < zbuffer[x0, yi0]:
                    if alpha[3] != 1.0:
                        talpha = image[x0, yi0, 3]
                        image[x0, yi0, 3] = alpha[3] + talpha * (1 - alpha[3])
                        for i in range(3):
                            image[x0, yi0, i] = (alpha[3]*alpha[i] + image[x0, yi0, i]*talpha*(1.0-alpha[3]))/image[x0,yi0,3]
                            if image[x0, yi0, 3] == 0.0:
                                image[x0, yi0, i] = 0.0
                    else:
                        for i in range(4):
                            image[x0, yi0, i] = alpha[i]
                    if (1.0 - image[x0, yi0, 3] < 1.0e-4):
                        zbuffer[x0, yi0] = z0
    return


def rotate_vectors(np.ndarray[np.float64_t, ndim=3] vecs,
        np.ndarray[np.float64_t, ndim=2] R):
    cdef int nx = vecs.shape[0]
    cdef int ny = vecs.shape[1]
    rotated = np.empty((nx,ny,3),dtype='float64')
    for i in range(nx):
        for j in range(ny):
            for k in range(3):
                rotated[i,j,k] =\
                    R[k,0]*vecs[i,j,0]+R[k,1]*vecs[i,j,1]+R[k,2]*vecs[i,j,2]
    return rotated

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def get_color_bounds(np.ndarray[np.float64_t, ndim=1] px,
                     np.ndarray[np.float64_t, ndim=1] py,
                     np.ndarray[np.float64_t, ndim=1] pdx,
                     np.ndarray[np.float64_t, ndim=1] pdy,
                     np.ndarray[np.float64_t, ndim=1] value,
                     np.float64_t leftx, np.float64_t rightx,
                     np.float64_t lefty, np.float64_t righty,
                     np.float64_t mindx = -1, np.float64_t maxdx = -1):
    cdef int i
    cdef np.float64_t mi = 1e100, ma = -1e100, v
    cdef int npx = px.shape[0]
    with nogil:
        for i in range(npx):
            v = value[i]
            if v < mi or v > ma:
                if px[i] + pdx[i] < leftx: continue
                if px[i] - pdx[i] > rightx: continue
                if py[i] + pdy[i] < lefty: continue
                if py[i] - pdy[i] > righty: continue
                if pdx[i] < mindx or pdy[i] < mindx: continue
                if maxdx > 0 and (pdx[i] > maxdx or pdy[i] > maxdx): continue
                if v < mi: mi = v
                if v > ma: ma = v
    return (mi, ma)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def kdtree_get_choices(np.ndarray[np.float64_t, ndim=3] data,
                       np.ndarray[np.float64_t, ndim=1] l_corner,
                       np.ndarray[np.float64_t, ndim=1] r_corner):
    cdef int i, j, k, dim, n_unique, best_dim, n_grids, my_split
    n_grids = data.shape[0]
    cdef np.float64_t **uniquedims
    cdef np.float64_t *uniques
    cdef np.float64_t split
    uniquedims = <np.float64_t **> alloca(3 * sizeof(np.float64_t*))
    for i in range(3):
        uniquedims[i] = <np.float64_t *> \
                alloca(2*n_grids * sizeof(np.float64_t))
    my_max = 0
    best_dim = -1
    my_split = -1
    for dim in range(3):
        n_unique = 0
        uniques = uniquedims[dim]
        for i in range(n_grids):
            # Check for disqualification
            for j in range(2):
                #print("Checking against", i,j,dim,data[i,j,dim])
                if not (l_corner[dim] < data[i, j, dim] and
                        data[i, j, dim] < r_corner[dim]):
                    #print("Skipping ", data[i,j,dim])
                    continue
                skipit = 0
                # Add our left ...
                for k in range(n_unique):
                    if uniques[k] == data[i, j, dim]:
                        skipit = 1
                        #print("Identified", uniques[k], data[i,j,dim], n_unique)
                        break
                if skipit == 0:
                    uniques[n_unique] = data[i, j, dim]
                    n_unique += 1
        if n_unique > my_max:
            best_dim = dim
            my_max = n_unique
            my_split = (n_unique-1)/2
    # I recognize how lame this is.
    cdef np.ndarray[np.float64_t, ndim=1] tarr = np.empty(my_max, dtype='float64')
    for i in range(my_max):
        #print("Setting tarr: ", i, uniquedims[best_dim][i])
        tarr[i] = uniquedims[best_dim][i]
    tarr.sort()
    if my_split < 0:
        raise RuntimeError
    split = tarr[my_split]
    cdef np.ndarray[np.uint8_t, ndim=1] less_ids = np.empty(n_grids, dtype='uint8')
    cdef np.ndarray[np.uint8_t, ndim=1] greater_ids = np.empty(n_grids, dtype='uint8')
    for i in range(n_grids):
        if data[i, 0, best_dim] < split:
            less_ids[i] = 1
        else:
            less_ids[i] = 0
        if data[i, 1, best_dim] > split:
            greater_ids[i] = 1
        else:
            greater_ids[i] = 0
    # Return out unique values
    return best_dim, split, less_ids.view("bool"), greater_ids.view("bool")

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def get_box_grids_level(np.ndarray[np.float64_t, ndim=1] left_edge,
                        np.ndarray[np.float64_t, ndim=1] right_edge,
                        int level,
                        np.ndarray[np.float64_t, ndim=2] left_edges,
                        np.ndarray[np.float64_t, ndim=2] right_edges,
                        np.ndarray[np.int32_t, ndim=2] levels,
                        np.ndarray[np.int32_t, ndim=1] mask,
                        int min_index = 0):
    cdef int i, n
    cdef int nx = left_edges.shape[0]
    cdef int inside
    cdef np.float64_t eps = np.finfo(np.float64).eps
    for i in range(nx):
        if i < min_index or levels[i,0] != level:
            mask[i] = 0
            continue
        inside = 1
        for n in range(3):
            if (right_edges[i,n] - left_edge[n]) <= eps or \
               (right_edge[n] - left_edges[i,n]) <= eps:
                inside = 0
                break
        if inside == 1: mask[i] = 1
        else: mask[i] = 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def get_box_grids_below_level(
                        np.ndarray[np.float64_t, ndim=1] left_edge,
                        np.ndarray[np.float64_t, ndim=1] right_edge,
                        int level,
                        np.ndarray[np.float64_t, ndim=2] left_edges,
                        np.ndarray[np.float64_t, ndim=2] right_edges,
                        np.ndarray[np.int32_t, ndim=2] levels,
                        np.ndarray[np.int32_t, ndim=1] mask,
                        int min_level = 0):
    cdef int i, n
    cdef int nx = left_edges.shape[0]
    cdef int inside
    cdef np.float64_t eps = np.finfo(np.float64).eps
    for i in range(nx):
        mask[i] = 0
        if levels[i,0] <= level and levels[i,0] >= min_level:
            inside = 1
            for n in range(3):
                if (right_edges[i,n] - left_edge[n]) <= eps or \
                   (right_edge[n] - left_edges[i,n]) <= eps:
                    inside = 0
                    break
            if inside == 1: mask[i] = 1

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def obtain_position_vector(
    data,
    field_names = (("index", "x"), ("index", "y"), ("index", "z"))
):
    # This is just to let the pointers exist and whatnot.  We can't cdef them
    # inside conditionals.
    cdef np.ndarray[np.float64_t, ndim=1] xf
    cdef np.ndarray[np.float64_t, ndim=1] yf
    cdef np.ndarray[np.float64_t, ndim=1] zf
    cdef np.ndarray[np.float64_t, ndim=2] rf
    cdef np.ndarray[np.float64_t, ndim=3] xg
    cdef np.ndarray[np.float64_t, ndim=3] yg
    cdef np.ndarray[np.float64_t, ndim=3] zg
    cdef np.ndarray[np.float64_t, ndim=4] rg
    cdef np.float64_t c[3]
    cdef int i, j, k

    units = data[field_names[0]].units
    center = data.get_field_parameter("center").to(units)
    c[0] = center[0]; c[1] = center[1]; c[2] = center[2]
    if len(data[field_names[0]].shape) == 1:
        # One dimensional data
        xf = data[field_names[0]]
        yf = data[field_names[1]]
        zf = data[field_names[2]]
        rf = YTArray(np.empty((3, xf.shape[0]), 'float64'), xf.units)
        for i in range(xf.shape[0]):
            rf[0, i] = xf[i] - c[0]
            rf[1, i] = yf[i] - c[1]
            rf[2, i] = zf[i] - c[2]
        return rf
    else:
        # Three dimensional data
        xg = data[field_names[0]]
        yg = data[field_names[1]]
        zg = data[field_names[2]]
        shape = (3, xg.shape[0], xg.shape[1], xg.shape[2])
        rg = YTArray(np.empty(shape, 'float64'), xg.units)
        #rg = YTArray(rg, xg.units)
        for i in range(xg.shape[0]):
            for j in range(xg.shape[1]):
                for k in range(xg.shape[2]):
                    rg[0,i,j,k] = xg[i,j,k] - c[0]
                    rg[1,i,j,k] = yg[i,j,k] - c[1]
                    rg[2,i,j,k] = zg[i,j,k] - c[2]
        return rg

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def obtain_relative_velocity_vector(
        data,
        field_names = (("gas", "velocity_x"), ("gas", "velocity_y"), ("gas", "velocity_z")),
        bulk_vector = "bulk_velocity"
    ):
    # This is just to let the pointers exist and whatnot.  We can't cdef them
    # inside conditionals.
    cdef np.ndarray[np.float64_t, ndim=1] vxf
    cdef np.ndarray[np.float64_t, ndim=1] vyf
    cdef np.ndarray[np.float64_t, ndim=1] vzf
    cdef np.ndarray[np.float64_t, ndim=2] rvf
    cdef np.ndarray[np.float64_t, ndim=3] vxg
    cdef np.ndarray[np.float64_t, ndim=3] vyg
    cdef np.ndarray[np.float64_t, ndim=3] vzg
    cdef np.ndarray[np.float64_t, ndim=4] rvg
    cdef np.float64_t bv[3]
    cdef int i, j, k, dim

    units = data[field_names[0]].units
    bulk_vector = data.get_field_parameter(bulk_vector).to(units)
    dim = data[field_names[0]].ndim
    if dim == 1:
        # One dimensional data
        vxf = data[field_names[0]].astype("float64")
        vyf = data[field_names[1]].astype("float64")
        vzf = data[field_names[2]].astype("float64")
        vyf.convert_to_units(vxf.units)
        vzf.convert_to_units(vxf.units)
        rvf = YTArray(np.empty((3, vxf.shape[0]), 'float64'), vxf.units)
        if bulk_vector is None:
            bv[0] = bv[1] = bv[2] = 0.0
        else:
            bulk_vector = bulk_vector.in_units(vxf.units)
            bv[0] = bulk_vector[0]
            bv[1] = bulk_vector[1]
            bv[2] = bulk_vector[2]
        for i in range(vxf.shape[0]):
            rvf[0, i] = vxf[i] - bv[0]
            rvf[1, i] = vyf[i] - bv[1]
            rvf[2, i] = vzf[i] - bv[2]
        return rvf
    elif dim == 3:
        # Three dimensional data
        vxg = data[field_names[0]].astype("float64")
        vyg = data[field_names[1]].astype("float64")
        vzg = data[field_names[2]].astype("float64")
        vyg.convert_to_units(vxg.units)
        vzg.convert_to_units(vxg.units)
        shape = (3, vxg.shape[0], vxg.shape[1], vxg.shape[2])
        rvg = YTArray(np.empty(shape, 'float64'), vxg.units)
        if bulk_vector is None:
            bv[0] = bv[1] = bv[2] = 0.0
        else:
            bulk_vector = bulk_vector.in_units(vxg.units)
            bv[0] = bulk_vector[0]
            bv[1] = bulk_vector[1]
            bv[2] = bulk_vector[2]
        for i in range(vxg.shape[0]):
            for j in range(vxg.shape[1]):
                for k in range(vxg.shape[2]):
                    rvg[0,i,j,k] = vxg[i,j,k] - bv[0]
                    rvg[1,i,j,k] = vyg[i,j,k] - bv[1]
                    rvg[2,i,j,k] = vzg[i,j,k] - bv[2]
        return rvg
    else:
        raise NotImplementedError(f"Unsupported dimensionality `{dim}`.")

def grow_flagging_field(oofield):
    cdef np.ndarray[np.uint8_t, ndim=3] ofield = oofield.astype("uint8")
    cdef np.ndarray[np.uint8_t, ndim=3] nfield
    nfield = np.zeros_like(ofield)
    cdef int i, j, k, ni, nj, nk
    cdef int oi, oj, ok
    for ni in range(ofield.shape[0]):
        for nj in range(ofield.shape[1]):
            for nk in range(ofield.shape[2]):
                for oi in range(3):
                    i = ni + (oi - 1)
                    if i < 0 or i >= ofield.shape[0]: continue
                    for oj in range(3):
                        j = nj + (oj - 1)
                        if j < 0 or j >= ofield.shape[1]: continue
                        for ok in range(3):
                            k = nk + (ok - 1)
                            if k < 0 or k >= ofield.shape[2]: continue
                            if ofield[i, j, k] == 1:
                                nfield[ni, nj, nk] = 1
    return nfield.astype("bool")

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def fill_region(
    input_fields,
    output_fields,
    np.int32_t output_level,
    np.ndarray[np.int64_t, ndim=1] left_index,
    np.ndarray[np.int64_t, ndim=2] ipos,
    np.ndarray[np.int64_t, ndim=1] ires,
    np.ndarray[np.int64_t, ndim=1] level_dims,
    np.ndarray[np.int64_t, ndim=1] refine_by
):
    cdef int i, n
    cdef np.int64_t tot = 0, oi, oj, ok
    cdef np.int64_t rf[3]
    cdef np.int64_t iind[3]
    cdef np.int64_t oind[3]
    cdef np.int64_t dim[3]
    cdef np.ndarray[np.float64_t, ndim=3] ofield
    cdef np.ndarray[np.float64_t, ndim=1] ifield
    nf = len(input_fields)
    # The variable offsets governs for each dimension and each possible
    # wrapping if we do it.  Then the wi, wj, wk indices check into each
    # [dim][wrap] inside the loops.
    cdef int wi, wj, wk
    cdef int offsets[3][3]
    cdef np.int64_t off
    for i in range(3):
        # Offsets here is a way of accounting for periodicity.  It keeps track
        # of how to offset our grid as we loop over the icoords.
        dim[i] = output_fields[0].shape[i]
        offsets[i][0] = offsets[i][2] = 0
        offsets[i][1] = 1
        if left_index[i] < 0:
            offsets[i][2] = 1
        if left_index[i] + dim[i] >= level_dims[i]:
            offsets[i][0] = 1
    for n in range(nf):
        tot = 0
        ofield = output_fields[n]
        ifield = input_fields[n]
        for i in range(ipos.shape[0]):
            for k in range(3):
                rf[k] = refine_by[k]**(output_level - ires[i])
            for wi in range(3):
                if offsets[0][wi] == 0: continue
                off = (left_index[0] + level_dims[0]*(wi-1))
                iind[0] = ipos[i, 0] * rf[0] - off
                # rf here is the "refinement factor", or, the number of zones
                # that this zone could potentially contribute to our filled
                # grid.
                for oi in range(rf[0]):
                    # Now we need to apply our offset
                    oind[0] = oi + iind[0]
                    if oind[0] < 0:
                        continue
                    elif oind[0] >= dim[0]:
                        break
                    for wj in range(3):
                        if offsets[1][wj] == 0: continue
                        off = (left_index[1] + level_dims[1]*(wj-1))
                        iind[1] = ipos[i, 1] * rf[1] - off
                        for oj in range(rf[1]):
                            oind[1] = oj + iind[1]
                            if oind[1] < 0:
                                continue
                            elif oind[1] >= dim[1]:
                                break
                            for wk in range(3):
                                if offsets[2][wk] == 0: continue
                                off = (left_index[2] + level_dims[2]*(wk-1))
                                iind[2] = ipos[i, 2] * rf[2] - off
                                for ok in range(rf[2]):
                                    oind[2] = ok + iind[2]
                                    if oind[2] < 0:
                                        continue
                                    elif oind[2] >= dim[2]:
                                        break
                                    ofield[oind[0], oind[1], oind[2]] = \
                                        ifield[i]
                                    tot += 1
    return tot

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def flip_bitmask(np.ndarray[np.float64_t, ndim=1] vals,
                 np.float64_t left_edge, np.float64_t right_edge,
                 np.uint64_t nbins):
    cdef np.uint64_t i, bin_id
    cdef np.float64_t idx = nbins / (right_edge - left_edge)
    cdef np.ndarray[np.uint8_t, ndim=1, cast=True] bitmask
    bitmask = np.zeros(nbins, dtype="uint8")
    for i in range(vals.shape[0]):
        bin_id = <np.uint64_t> ((vals[i] - left_edge)*idx)
        bitmask[bin_id] = 1
    return bitmask

#@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def flip_morton_bitmask(np.ndarray[np.uint64_t, ndim=1] morton_indices,
                        int max_order):
    # We assume that the morton_indices are fed to us in a setup that allows
    # for 20 levels.  This means that we shift right by 3*(20-max_order) (or is
    # that a fencepost?)
    cdef np.uint64_t mi, i
    cdef np.ndarray[np.uint8_t, ndim=1, cast=True] bitmask
    # Note that this will fail if it's too big, since numpy will check nicely
    # the memory availability.  I guess.
    bitmask = np.zeros(1 << (3*max_order), dtype="uint8")
    for i in range(morton_indices.shape[0]):
        mi = (morton_indices[i] >> (3 * (20-max_order)))
        bitmask[mi] = 1
    return bitmask

#@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def count_collisions(np.ndarray[np.uint8_t, ndim=2] masks):
    cdef int i, j, k
    cdef np.ndarray[np.uint32_t, ndim=1] counts
    cdef np.ndarray[np.uint8_t, ndim=1] collides
    counts = np.zeros(masks.shape[1], dtype="uint32")
    collides = np.zeros(masks.shape[1], dtype="uint8")
    for i in range(masks.shape[1]):
        print(i)
        for j in range(masks.shape[1]):
            collides[j] = 0
        for k in range(masks.shape[0]):
            if masks[k,i] == 0: continue
            for j in range(masks.shape[1]):
                if j == i: continue
                if masks[k,j] == 1:
                    collides[j] = 1
        counts[i] = collides.sum()
    return counts

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def fill_region_float(np.ndarray[np.float64_t, ndim=2] fcoords,
                      np.ndarray[np.float64_t, ndim=2] fwidth,
                      np.ndarray[np.float64_t, ndim=1] data,
                      np.ndarray[np.float64_t, ndim=1] box_left_edge,
                      np.ndarray[np.float64_t, ndim=1] box_right_edge,
                      np.ndarray[np.float64_t, ndim=3] dest,
                      int antialias = 1,
                      period = None,
                      int check_period = 1):
    cdef np.float64_t ds_period[3]
    cdef np.float64_t box_dds[3]
    cdef np.float64_t box_idds[3]
    cdef np.float64_t width[3]
    cdef np.float64_t LE[3]
    cdef np.float64_t RE[3]
    cdef np.int64_t i, j, k, p, xi, yi
    cdef np.int64_t dims[3]
    cdef np.int64_t ld[3]
    cdef np.int64_t ud[3]
    cdef np.float64_t overlap[3]
    cdef np.float64_t dsp
    cdef np.float64_t osp[3]
    cdef np.float64_t odsp[3]
    cdef np.float64_t sp[3]
    cdef np.float64_t lfd[3]
    cdef np.float64_t ufd[3]
    # These are the temp vars we get from the arrays
    # Some periodicity helpers
    cdef int diter[3][2]
    cdef np.float64_t diterv[3][2]
    if period is not None:
        for i in range(3):
            ds_period[i] = period[i]
    else:
        ds_period[0] = ds_period[1] = ds_period[2] = 0.0
    box_left_edge = _ensure_code(box_left_edge)
    box_right_edge = _ensure_code(box_right_edge)
    _ensure_code(fcoords)
    _ensure_code(fwidth)
    for i in range(3):
        LE[i] = box_left_edge[i]
        RE[i] = box_right_edge[i]
        width[i] = RE[i] - LE[i]
        dims[i] = dest.shape[i]
        box_dds[i] = width[i] / dims[i]
        box_idds[i] = 1.0/box_dds[i]
        diter[i][0] = diter[i][1] = 0
        diterv[i][0] = diterv[i][1] = 0.0
        overlap[i] = 1.0
    with nogil:
        for p in range(fcoords.shape[0]):
            for i in range(3):
               diter[i][1] = 999
               odsp[i] = fwidth[p,i]*0.5
               osp[i] = fcoords[p,i] # already centered
               overlap[i] = 1.0
            dsp = data[p]
            if check_period == 1:
                for i in range(3):
                    if (osp[i] - odsp[i] < LE[i]):
                        diter[i][1] = +1
                        diterv[i][1] = ds_period[i]
                    elif (osp[i] + odsp[i] > RE[i]):
                        diter[i][1] = -1
                        diterv[i][1] = -ds_period[i]
            for xi in range(2):
                if diter[0][xi] == 999: continue
                sp[0] = osp[0] + diterv[0][xi]
                if (sp[0] + odsp[0] < LE[0]) or (sp[0] - odsp[0] > RE[0]): continue
                for yi in range(2):
                    if diter[1][yi] == 999: continue
                    sp[1] = osp[1] + diterv[1][yi]
                    if (sp[1] + odsp[1] < LE[1]) or (sp[1] - odsp[1] > RE[1]): continue
                    for zi in range(2):
                        if diter[2][zi] == 999: continue
                        sp[2] = osp[2] + diterv[2][yi]
                        if (sp[2] + odsp[2] < LE[2]) or (sp[2] - odsp[2] > RE[2]): continue
                        for i in range(3):
                            ld[i] = <np.int64_t> fmax(((sp[i]-odsp[i]-LE[i])*box_idds[i]),0)
                            # NOTE: This is a different way of doing it than in the C
                            # routines.  In C, we were implicitly casting the
                            # initialization to int, but *not* the conditional, which
                            # was allowed an extra value:
                            #     for(j=lc;j<rc;j++)
                            # here, when assigning lc (double) to j (int) it got
                            # truncated, but no similar truncation was done in the
                            # comparison of j to rc (double).  So give ourselves a
                            # bonus row and bonus column here.
                            ud[i] = <np.int64_t> fmin(((sp[i]+odsp[i]-LE[i])*box_idds[i] + 1), dims[i])
                        for i in range(ld[0], ud[0]):
                            if antialias == 1:
                                lfd[0] = box_dds[0] * i + LE[0]
                                ufd[0] = box_dds[0] * (i + 1) + LE[0]
                                overlap[0] = ((fmin(ufd[0], sp[0]+odsp[0])
                                           - fmax(lfd[0], (sp[0]-odsp[0])))*box_idds[0])
                            if overlap[0] < 0.0: continue
                            for j in range(ld[1], ud[1]):
                                if antialias == 1:
                                    lfd[1] = box_dds[1] * j + LE[1]
                                    ufd[1] = box_dds[1] * (j + 1) + LE[1]
                                    overlap[1] = ((fmin(ufd[1], sp[1]+odsp[1])
                                               - fmax(lfd[1], (sp[1]-odsp[1])))*box_idds[1])
                                if overlap[1] < 0.0: continue
                                for k in range(ld[2], ud[2]):
                                    if antialias == 1:
                                        lfd[2] = box_dds[2] * k + LE[2]
                                        ufd[2] = box_dds[2] * (k + 1) + LE[2]
                                        overlap[2] = ((fmin(ufd[2], sp[2]+odsp[2])
                                                   - fmax(lfd[2], (sp[2]-odsp[2])))*box_idds[2])
                                        if overlap[2] < 0.0: continue
                                        dest[i,j,k] += dsp * (overlap[0]*overlap[1]*overlap[2])
                                    else:
                                        dest[i,j,k] = dsp

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def gravitational_binding_energy(
        np.float64_t[:] mass,
        np.float64_t[:] x,
        np.float64_t[:] y,
        np.float64_t[:] z,
        int truncate,
        np.float64_t kinetic,
        int num_threads = 0):

    cdef int q_outer, q_inner, n_q
    cdef np.float64_t mass_o, x_o, y_o, z_o
    cdef np.float64_t mass_i, x_i, y_i, z_i
    cdef np.float64_t total_potential = 0.
    cdef np.float64_t this_potential

    n_q = mass.size
    pbar = get_pbar("Calculating potential for %d cells with %d thread(s)" % (n_q,num_threads),
        n_q)

    # using reversed iterator in order to make use of guided scheduling
    # (inner loop is getting more and more expensive)
    for q_outer in prange(n_q - 1,-1,-1,
        nogil=True,schedule='guided',num_threads=num_threads):
        this_potential = 0.

        mass_o = mass[q_outer]
        x_o = x[q_outer]
        y_o = y[q_outer]
        z_o = z[q_outer]
        for q_inner in range(q_outer + 1, n_q):
            mass_i = mass[q_inner]
            x_i = x[q_inner]
            y_i = y[q_inner]
            z_i = z[q_inner]
            # not using += operator so that variable is not automatically reduced
            this_potential = this_potential + mass_o * mass_i / \
              sqrt((x_i - x_o) * (x_i - x_o) +
                   (y_i - y_o) * (y_i - y_o) +
                   (z_i - z_o) * (z_i - z_o))
        total_potential += this_potential
        if truncate and this_potential / kinetic > 1.:
            break
        with gil:
            PyErr_CheckSignals()
            # this call is not thread safe, but it gives a reasonable approximation
            pbar.update()

    pbar.finish()
    return total_potential

# The OnceIndirect code is from:
# http://stackoverflow.com/questions/10465091/assembling-a-cython-memoryview-from-numpy-arrays/12991519#12991519
# This is under the CC-BY-SA license.

cdef class OnceIndirect:
    cdef object _objects
    cdef void** buf
    cdef int ndim
    cdef int n_rows
    cdef int buf_len
    cdef Py_ssize_t* shape
    cdef Py_ssize_t* strides
    cdef Py_ssize_t* suboffsets
    cdef Py_ssize_t itemsize
    cdef bytes format
    cdef int is_readonly

    def __cinit__(self, object rows, want_writable=True, want_format=True, allow_indirect=False):
        """
        Set want_writable to False if you don't want writable data. (This may
        prevent copies.)
        Set want_format to False if your input doesn't support PyBUF_FORMAT (unlikely)
        Set allow_indirect to True if you are ok with the memoryview being indirect
        in dimensions other than the first. (This may prevent copies.)

        An example usage:

        cdef double[::cython.view.indirect, ::1] vectors =
            OnceIndirect([object.vector for object in objects])
        """
        demand = buffer.PyBUF_INDIRECT if allow_indirect else buffer.PyBUF_STRIDES
        if want_writable:
            demand |= buffer.PyBUF_WRITABLE
        if want_format:
            demand |= buffer.PyBUF_FORMAT
        self._objects = [memoryview(row, demand) for row in rows]
        self.n_rows = len(self._objects)
        self.buf_len = sizeof(void*) * self.n_rows
        self.buf = <void**>malloc(self.buf_len)
        self.ndim = 1 + self._objects[0].ndim
        self.shape = <Py_ssize_t*>malloc(sizeof(Py_ssize_t) * self.ndim)
        self.strides = <Py_ssize_t*>malloc(sizeof(Py_ssize_t) * self.ndim)
        self.suboffsets = <Py_ssize_t*>malloc(sizeof(Py_ssize_t) * self.ndim)

        cdef memoryview example_obj = self._objects[0]
        self.itemsize = example_obj.itemsize

        if want_format:
            self.format = example_obj.view.format
        else:
            self.format = None
        self.is_readonly |= example_obj.view.readonly

        for dim in range(self.ndim):
            if dim == 0:
                self.shape[dim] = self.n_rows
                self.strides[dim] = sizeof(void*)
                self.suboffsets[dim] = 0
            else:
                self.shape[dim] = example_obj.view.shape[dim - 1]
                self.strides[dim] = example_obj.view.strides[dim - 1]
                if example_obj.view.suboffsets == NULL:
                    self.suboffsets[dim] = -1
                else:
                    self.suboffsets[dim] = example_obj.suboffsets[dim - 1]

        cdef memoryview obj
        cdef int i = 0
        for obj in self._objects:
            assert_similar(example_obj, obj)
            self.buf[i] = obj.view.buf
            i += 1

    def __getbuffer__(self, Py_buffer* buff, int flags):
        if (flags & buffer.PyBUF_INDIRECT) != buffer.PyBUF_INDIRECT:
            raise Exception("don't want to copy data")
        if flags & buffer.PyBUF_WRITABLE and self.is_readonly:
            raise Exception("couldn't provide writable, you should have demanded it earlier")
        if flags & buffer.PyBUF_FORMAT:
            if self.format is None:
                raise Exception("couldn't provide format, you should have demanded it earlier")
            buff.format = self.format
        else:
            buff.format = NULL

        buff.buf = <void*>self.buf
        buff.obj = self
        buff.len = self.buf_len
        buff.readonly = self.is_readonly
        buff.ndim = self.ndim
        buff.shape = self.shape
        buff.strides = self.strides
        buff.suboffsets = self.suboffsets
        buff.itemsize = self.itemsize
        buff.internal = NULL

    def __dealloc__(self):
        free(self.buf)
        free(self.shape)
        free(self.strides)
        free(self.suboffsets)

cdef int assert_similar(memoryview left_, memoryview right_) except -1:
    cdef Py_buffer left = left_.view
    cdef Py_buffer right = right_.view
    assert left.ndim == right.ndim
    cdef int i
    for i in range(left.ndim):
        assert left.shape[i] == right.shape[i], (left_.shape, right_.shape)
        assert left.strides[i] == right.strides[i], (left_.strides, right_.strides)

    if left.suboffsets == NULL:
        assert right.suboffsets == NULL, (left_.suboffsets, right_.suboffsets)
    else:
        for i in range(left.ndim):
            assert left.suboffsets[i] == right.suboffsets[i], (left_.suboffsets, right_.suboffsets)

    if left.format == NULL:
        assert right.format == NULL, (bytes(left.format), bytes(right.format))
    else:
        #alternatively, compare as Python strings:
        #assert bytes(left.format) == bytes(right.format)
        assert strcmp(left.format, right.format) == 0, (bytes(left.format), bytes(right.format))
    return 0

def _obtain_coords_and_widths(np.int64_t[:] icoords,
                              np.int64_t[:] ires,
                              np.float64_t[:] cell_widths,
                              np.float64_t offset):
    # This function only accepts *one* axis of icoords, because we will be
    # looping over the axes in python.  This also simplifies cell_width
    # allocation and computation.
    cdef np.ndarray[np.float64_t, ndim=1] fcoords = np.zeros(icoords.size, dtype="f8")
    cdef np.ndarray[np.float64_t, ndim=1] fwidth = np.zeros(icoords.size, dtype="f8")
    cdef np.ndarray[np.float64_t, ndim=1] cell_centers = np.zeros(cell_widths.size, dtype="f8")
    cdef int i
    cdef np.float64_t pos = offset
    for i in range(cell_widths.size):
        cell_centers[i] = pos + 0.5 * cell_widths[i]
        pos += cell_widths[i]
    for i in range(icoords.size):
        fcoords[i] = cell_centers[icoords[i]]
        fwidth[i] = cell_widths[icoords[i]]
    return fcoords, fwidth



from itertools import product

import numpy as np

from yt.funcs import mylog
from yt.utilities.lib._octree_raytracing import _OctreeRayTracing


class OctreeRayTracing:
    octree = None
    data_source = None
    log_fields = None
    fields = None

    # Internal data
    _cell_index = None
    _tvalues = None

    def __init__(self, data_source):
        self.data_source = data_source
        ds = data_source.ds
        LE = np.array([0, 0, 0], dtype=np.float64)
        RE = np.array([1, 1, 1], dtype=np.float64)
        lvl_min = ds.min_level + 1
        # This is the max refinement so that the smallest cells have size
        # 1/2**depth
        depth = lvl_min + ds.max_level + 1

        self.octree = _OctreeRayTracing(LE, RE, depth)
        ds = data_source.ds

        xyz = np.stack([data_source[key].to("unitary").value for key in "xyz"], axis=-1)
        lvl = data_source["grid_level"].astype(int).value + lvl_min

        ipos = np.floor(xyz * (1 << depth)).astype(int)
        mylog.debug("Adding cells to volume")
        self.octree.add_nodes(
            ipos.astype(np.int32),
            lvl.astype(np.int32),
            np.arange(len(ipos), dtype=np.int32),
        )

    def vertex_centered_data(self, field):
        data_source = self.data_source
        chunks = data_source.index._chunk(data_source, "spatial", ngz=1)

        finfo = data_source.ds._get_field_info(field)
        units = finfo.units
        rv = data_source.ds.arr(
            np.zeros((2, 2, 2, data_source.ires.size), dtype="float64"), units
        )
        binary_3D_index_iter = product(*[range(2)] * 3)
        ind = {(i, j, k): 0 for i, j, k in binary_3D_index_iter}
        for chunk in chunks:
            with data_source._chunked_read(chunk):
                gz = data_source._current_chunk.objs[0]
                gz.field_parameters = data_source.field_parameters
                wogz = gz._base_grid
                vertex_data = gz.get_vertex_centered_data([field])[field]

                for i, j, k in product(*[range(2)] * 3):
                    ind[i, j, k] += wogz.select(
                        data_source.selector,
                        vertex_data[i : i + 2, j : j + 2, k : k + 2, ...],
                        rv[i, j, k, :],
                        ind[i, j, k],
                    )
        return rv

    def set_fields(self, fields, log_fields, no_ghost, force=False):
        if no_ghost:
            raise NotImplementedError("Ghost zones are required with Octree datasets")

        if len(fields) != 1:
            raise ValueError(
                "Can only set one fields at a time. "
                "This is likely a bug, and should be reported."
            )

        field = self.data_source._determine_fields(fields)[0]
        take_log = log_fields[0]
        vertex_data = self.vertex_centered_data(field)

        if take_log:
            vertex_data = np.log10(vertex_data)

        # Vertex_data has shape (2, 2, 2, ...)
        # Note: here we have the wrong ordering within the oct (classical Fortran/C
        # ordering issue) so we need to swap axis 0 and 2.
        self.data = vertex_data.swapaxes(0, 2).reshape(8, -1)

    def cast_rays(self, vp_pos, vp_dir):
        """Cast the rays through the oct.

        Parameters
        ----------
        vp_pos : float arrays (Nrays, Ndim)
        vp_dir : float arrays (Nrays, Ndim)
            The position (unitary) and direction of each ray

        Returns
        -------
        cell_index : list of integer arrays of shape (Ncell)
            For each ray, contains an ordered array of cell ids
            that it intersects with
        tvalues : list of float arrays of shape (Ncell, 2)
            The t value at entry and exit for each cell.
        """
        if not self._cell_index:
            self._cell_index, self._tvalues = self.octree.cast_rays(vp_pos, vp_dir)
        return self._cell_index, self._tvalues



from numpy.testing import assert_array_equal, assert_equal

from yt.testing import fake_random_ds


def test_clone_sphere():
    # Now we test that we can get different radial velocities based on field
    # parameters.
    fields = ("density", "velocity_x", "velocity_y", "velocity_z")
    units = ("g/cm**3", "cm/s", "cm/s", "cm/s")
    # Get the first sphere
    ds = fake_random_ds(16, fields=fields, units=units)
    sp0 = ds.sphere(ds.domain_center, 0.25)

    assert_equal(list(sp0.keys()), [])

    sp1 = sp0.clone()
    sp0[("gas", "density")]
    assert_equal(list(sp0.keys()), (("gas", "density"),))
    assert_equal(list(sp1.keys()), [])

    sp1[("gas", "density")]

    assert_array_equal(sp0[("gas", "density")], sp1[("gas", "density")])


def test_clone_cut_region():
    fields = ("density", "temperature")
    units = ("g/cm**3", "K")
    ds = fake_random_ds(64, nprocs=4, fields=fields, units=units)
    dd = ds.all_data()
    reg1 = dd.cut_region(
        ["obj[('gas', 'temperature')] > 0.5", "obj[('gas', 'density')] < 0.75"]
    )
    reg2 = reg1.clone()
    assert_array_equal(reg1[("gas", "density")], reg2[("gas", "density")])



import numpy as np

import yt
from yt.data_objects.level_sets.api import Clump, find_clumps

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

data_source = ds.disk([0.5, 0.5, 0.5], [0.0, 0.0, 1.0], (8, "kpc"), (1, "kpc"))

# the field to be used for contouring
field = ("gas", "density")

# This is the multiplicative interval between contours.
step = 2.0

# Now we set some sane min/max values between which we want to find contours.
# This is how we tell the clump finder what to look for -- it won't look for
# contours connected below or above these threshold values.
c_min = 10 ** np.floor(np.log10(data_source[field]).min())
c_max = 10 ** np.floor(np.log10(data_source[field]).max() + 1)

# Now find get our 'base' clump -- this one just covers the whole domain.
master_clump = Clump(data_source, field)

# Add a "validator" to weed out clumps with less than 20 cells.
# As many validators can be added as you want.
master_clump.add_validator("min_cells", 20)

# Calculate center of mass for all clumps.
master_clump.add_info_item("center_of_mass")

# Begin clump finding.
find_clumps(master_clump, c_min, c_max, step)

# Save the clump tree as a reloadable dataset
fn = master_clump.save_as_dataset(fields=[("gas", "density"), ("all", "particle_mass")])

# We can traverse the clump hierarchy to get a list of all of the 'leaf' clumps
leaf_clumps = master_clump.leaves

# Get total cell and particle masses for each leaf clump
leaf_masses = [leaf.quantities.total_mass() for leaf in leaf_clumps]

# If you'd like to visualize these clumps, a list of clumps can be supplied to
# the "clumps" callback on a plot.  First, we create a projection plot:
prj = yt.ProjectionPlot(ds, 2, field, center="c", width=(20, "kpc"))

# Next we annotate the plot with contours on the borders of the clumps
prj.annotate_clumps(leaf_clumps)

# Save the plot to disk.
prj.save("clumps")

# Reload the clump dataset.
cds = yt.load(fn)

# Clump annotation can also be done with the reloaded clump dataset.

# Remove the original clump annotation
prj.clear_annotations()

# Get the leaves and add the callback.
leaf_clumps_reloaded = cds.leaves
prj.annotate_clumps(leaf_clumps_reloaded)
prj.save("clumps_reloaded")

# Query fields for clumps in the tree.
print(cds.tree["clump", "center_of_mass"])
print(cds.tree.children[0]["grid", "density"])
print(cds.tree.children[1]["all", "particle_mass"])

# Get all of the leaf clumps.
print(cds.leaves)
print(cds.leaves[0]["clump", "cell_mass"])



"""
ytdata frontend tests using enzo_tiny_cosmology



"""

import os
import shutil
import tempfile

import numpy as np

from yt.data_objects.api import create_profile
from yt.frontends.ytdata.api import (
    YTDataContainerDataset,
    YTGridDataset,
    YTNonspatialDataset,
    YTProfileDataset,
    YTSpatialPlotDataset,
)
from yt.frontends.ytdata.tests.test_outputs import (
    YTDataFieldTest,
    compare_unit_attributes,
)
from yt.testing import (
    assert_allclose_units,
    assert_array_equal,
    requires_file,
    requires_module,
    skip,
)
from yt.units.yt_array import YTArray
from yt.utilities.answer_testing.framework import data_dir_load, requires_ds
from yt.visualization.profile_plotter import PhasePlot, ProfilePlot

enzotiny = "enzo_tiny_cosmology/DD0046/DD0046"
ytdata_dir = "ytdata_test"


@skip(reason="See https://github.com/yt-project/yt/issues/3909")
@requires_module("h5py")
@requires_ds(enzotiny)
@requires_file(os.path.join(ytdata_dir, "DD0046_sphere.h5"))
@requires_file(os.path.join(ytdata_dir, "DD0046_cut_region.h5"))
def test_old_datacontainer_data():
    ds = data_dir_load(enzotiny)
    sphere = ds.sphere(ds.domain_center, (10, "Mpc"))
    fn = "DD0046_sphere.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    sphere_ds = data_dir_load(full_fn)
    compare_unit_attributes(ds, sphere_ds)
    assert isinstance(sphere_ds, YTDataContainerDataset)
    yield YTDataFieldTest(full_fn, ("grid", "density"))
    yield YTDataFieldTest(full_fn, ("all", "particle_mass"))
    cr = ds.cut_region(sphere, ['obj[("gas", "temperature")] > 1e4'])
    fn = "DD0046_cut_region.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    cr_ds = data_dir_load(full_fn)
    assert isinstance(cr_ds, YTDataContainerDataset)
    assert (cr[("gas", "temperature")] == cr_ds.data[("gas", "temperature")]).all()


@skip(reason="See https://github.com/yt-project/yt/issues/3909")
@requires_module("h5py")
@requires_ds(enzotiny)
@requires_file(os.path.join(ytdata_dir, "DD0046_covering_grid.h5"))
@requires_file(os.path.join(ytdata_dir, "DD0046_arbitrary_grid.h5"))
@requires_file(os.path.join(ytdata_dir, "DD0046_proj_frb.h5"))
def test_old_grid_datacontainer_data():
    ds = data_dir_load(enzotiny)

    fn = "DD0046_covering_grid.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    cg_ds = data_dir_load(full_fn)
    compare_unit_attributes(ds, cg_ds)
    assert isinstance(cg_ds, YTGridDataset)
    yield YTDataFieldTest(full_fn, ("grid", "density"))
    yield YTDataFieldTest(full_fn, ("all", "particle_mass"))

    fn = "DD0046_arbitrary_grid.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    ag_ds = data_dir_load(full_fn)
    compare_unit_attributes(ds, ag_ds)
    assert isinstance(ag_ds, YTGridDataset)
    yield YTDataFieldTest(full_fn, ("grid", "density"))
    yield YTDataFieldTest(full_fn, ("all", "particle_mass"))

    my_proj = ds.proj("density", "x", weight_field="density")
    frb = my_proj.to_frb(1.0, (800, 800))
    fn = "DD0046_proj_frb.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    frb_ds = data_dir_load(full_fn)
    assert_allclose_units(
        frb[("gas", "density")], frb_ds.data[("gas", "density")], 1e-7
    )
    compare_unit_attributes(ds, frb_ds)
    assert isinstance(frb_ds, YTGridDataset)
    yield YTDataFieldTest(full_fn, ("gas", "density"), geometric=False)


@skip(reason="See https://github.com/yt-project/yt/issues/3909")
@requires_module("h5py")
@requires_ds(enzotiny)
@requires_file(os.path.join(ytdata_dir, "DD0046_proj.h5"))
def test_old_spatial_data():
    ds = data_dir_load(enzotiny)
    fn = "DD0046_proj.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    proj_ds = data_dir_load(full_fn)
    compare_unit_attributes(ds, proj_ds)
    assert isinstance(proj_ds, YTSpatialPlotDataset)
    yield YTDataFieldTest(full_fn, ("gas", "density"), geometric=False)


@skip(reason="See https://github.com/yt-project/yt/issues/3909")
@requires_module("h5py")
@requires_ds(enzotiny)
@requires_file(os.path.join(ytdata_dir, "DD0046_Profile1D.h5"))
@requires_file(os.path.join(ytdata_dir, "DD0046_Profile2D.h5"))
def test_old_profile_data():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    ds = data_dir_load(enzotiny)
    ad = ds.all_data()
    profile_1d = create_profile(
        ad,
        ("gas", "density"),
        ("gas", "temperature"),
        weight_field=("gas", "cell_mass"),
    )
    fn = "DD0046_Profile1D.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    prof_1d_ds = data_dir_load(full_fn)
    compare_unit_attributes(ds, prof_1d_ds)
    assert isinstance(prof_1d_ds, YTProfileDataset)

    for field in profile_1d.standard_deviation:
        assert_array_equal(
            profile_1d.standard_deviation[field],
            prof_1d_ds.profile.standard_deviation["data", field[1]],
        )

    p1 = ProfilePlot(
        prof_1d_ds.data,
        ("gas", "density"),
        ("gas", "temperature"),
        weight_field=("gas", "cell_mass"),
    )
    p1.save()

    yield YTDataFieldTest(full_fn, ("gas", "temperature"), geometric=False)
    yield YTDataFieldTest(full_fn, ("index", "x"), geometric=False)
    yield YTDataFieldTest(full_fn, ("gas", "density"), geometric=False)
    fn = "DD0046_Profile2D.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    prof_2d_ds = data_dir_load(full_fn)
    compare_unit_attributes(ds, prof_2d_ds)
    assert isinstance(prof_2d_ds, YTProfileDataset)

    p2 = PhasePlot(
        prof_2d_ds.data,
        ("gas", "density"),
        ("gas", "temperature"),
        ("gas", "cell_mass"),
        weight_field=None,
    )
    p2.save()

    yield YTDataFieldTest(full_fn, ("gas", "density"), geometric=False)
    yield YTDataFieldTest(full_fn, ("index", "x"), geometric=False)
    yield YTDataFieldTest(full_fn, ("gas", "temperature"), geometric=False)
    yield YTDataFieldTest(full_fn, ("index", "y"), geometric=False)
    yield YTDataFieldTest(full_fn, ("gas", "cell_mass"), geometric=False)
    os.chdir(curdir)
    shutil.rmtree(tmpdir)


@skip(reason="See https://github.com/yt-project/yt/issues/3909")
@requires_module("h5py")
@requires_ds(enzotiny)
@requires_file(os.path.join(ytdata_dir, "test_data.h5"))
@requires_file(os.path.join(ytdata_dir, "random_data.h5"))
def test_old_nonspatial_data():
    ds = data_dir_load(enzotiny)
    region = ds.box([0.25] * 3, [0.75] * 3)
    sphere = ds.sphere(ds.domain_center, (10, "Mpc"))
    my_data = {}
    my_data["region_density"] = region[("gas", "density")]
    my_data["sphere_density"] = sphere[("gas", "density")]
    fn = "test_data.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    array_ds = data_dir_load(full_fn)
    compare_unit_attributes(ds, array_ds)
    assert isinstance(array_ds, YTNonspatialDataset)
    yield YTDataFieldTest(full_fn, "region_density", geometric=False)
    yield YTDataFieldTest(full_fn, "sphere_density", geometric=False)

    my_data = {"density": YTArray(np.linspace(1.0, 20.0, 10), "g/cm**3")}
    fn = "random_data.h5"
    full_fn = os.path.join(ytdata_dir, fn)
    new_ds = data_dir_load(full_fn)
    assert isinstance(new_ds, YTNonspatialDataset)
    yield YTDataFieldTest(full_fn, ("gas", "density"), geometric=False)



import yt

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# The maximum refinement level of this dataset is 8
print(ds.max_level)

# If we ask for *all* of the AMR data, we get back field
# values sampled at about 3.6 million AMR zones
ad = ds.all_data()
print(ad["gas", "density"].shape)

# Let's only sample data up to AMR level 2
ad.max_level = 2

# Now we only sample from about 200,000 zones
print(ad["gas", "density"].shape)

# Note that this includes data at level 2 that would
# normally be masked out. There aren't any "holes" in
# the downsampled AMR mesh, the volume still sums to
# the volume of the domain:
print(ad["gas", "volume"].sum())
print(ds.domain_width.prod())

# Now let's make a downsampled plot
plot = yt.SlicePlot(ds, "z", ("gas", "density"), data_source=ad)
plot.save("downsampled.png")
